from collections import deque
from collections import OrderedDict
from difflib import get_close_matches
import functools
import inspect
from itertools import chain
import math
from warnings import warn

import numpy as np
import tensorflow as tf
from tensorflow.python.training import moving_averages


class G:
    pass


G._n_layers = 0


def create_param(spec, shape, name, trainable=True, regularizable=True):
    if not hasattr(spec, '__call__'):
        assert isinstance(spec, (tf.Tensor, tf.Variable))
        return spec
    assert hasattr(spec, '__call__')
    if regularizable:
        # use the default regularizer
        regularizer = None
    else:
        # do not regularize this variable
        def f(_):
            return tf.constant(0.)

        regularizer = f
    return tf.compat.v1.get_variable(
        name=name,
        shape=shape,
        initializer=spec,
        trainable=trainable,
        regularizer=regularizer,
        dtype=tf.float32)


def as_tuple(x, n, t=None):
    try:
        x = tuple(x)
    except TypeError:
        x = (x, ) * n

    if (t is not None) and not all(isinstance(v, t) for v in x):
        raise TypeError('expected a single value or an iterable '
                        'of {0}, got {1} instead'.format(t.__name__, x))

    if len(x) != n:
        raise ValueError('expected a single value or an iterable '
                         'with length {0}, got {1} instead'.format(n, x))

    return x


def conv_output_length(input_length, filter_size, stride, pad=0):
    """Helper function to compute the output size of a convolution operation
    This function computes the length along a single axis, which corresponds
    to a 1D convolution. It can also be used for convolutions with higher
    dimensionalities by using it individually for each axis.
    Parameters
    ----------
    input_length : int or None
        The size of the input.
    filter_size : int
        The size of the filter.
    stride : int
        The stride of the convolution operation.
    pad : int, 'full' or 'same' (default: 0)
        By default, the convolution is only computed where the input and the
        filter fully overlap (a valid convolution). When ``stride=1``, this
        yields an output that is smaller than the input by ``filter_size - 1``.
        The `pad` argument allows you to implicitly pad the input with zeros,
        extending the output size.
        A single integer results in symmetric zero-padding of the given size on
        both borders.
        ``'full'`` pads with one less than the filter size on both sides. This
        is equivalent to computing the convolution wherever the input and the
        filter overlap by at least one position.
        ``'same'`` pads with half the filter size on both sides (one less on
        the second side for an even filter size). When ``stride=1``, this
        results in an output size equal to the input size.
    Returns
    -------
    int or None
        The output size corresponding to the given convolution parameters, or
        ``None`` if `input_size` is ``None``.
    Raises
    ------
    ValueError
        When an invalid padding is specified, a `ValueError` is raised.
    """
    if input_length is None:
        return None
    if pad == 'valid':
        output_length = input_length - filter_size + 1
    elif pad == 'full':
        output_length = input_length + filter_size - 1
    elif pad == 'same':
        output_length = input_length
    elif isinstance(pad, int):
        output_length = input_length + 2 * pad - filter_size + 1
    else:
        raise ValueError('Invalid pad: {0}'.format(pad))

    # This is the integer arithmetic equivalent to
    # np.ceil(output_length / stride)
    output_length = (output_length + stride - 1) // stride

    return output_length


class Layer:
    def __init__(self,
                 incoming,
                 name=None,
                 variable_reuse=None,
                 weight_normalization=False,
                 **kwargs):
        if isinstance(incoming, tuple):
            self.input_shape = incoming
            self.input_layer = None
        else:
            self.input_shape = incoming.output_shape
            self.input_layer = incoming
        self.params = OrderedDict()
        self.weight_normalization = weight_normalization

        if name is None:
            name = '%s_%d' % (type(self).__name__, G._n_layers)
            G._n_layers += 1

        self.name = name
        self.variable_reuse = variable_reuse
        self.get_output_kwargs = []
        self.variable_scope = tf.compat.v1.variable_scope(
            name, reuse=self.variable_reuse)

        if any(d is not None and d <= 0 for d in self.input_shape):
            raise ValueError(
                ('Cannot create Layer with a non-positive input_shape '
                 'dimension. input_shape=%r, self.name=%r') %
                (self.input_shape, self.name))

    @property
    def output_shape(self):
        shape = self.get_output_shape_for(self.input_shape)
        if any(isinstance(s, (tf.Variable, tf.Tensor)) for s in shape):
            raise ValueError('%s returned a symbolic output shape from its '
                             'get_output_shape_for() method: %r. This is not '
                             'allowed; shapes must be tuples of integers for '
                             'fixed-size dimensions and Nones for variable '
                             'dimensions.' % (self.__class__.__name__, shape))
        return shape

    def get_output_shape_for(self, input_shape):
        raise NotImplementedError

    def get_output_for(self, input, **kwargs):
        raise NotImplementedError

    def add_param_plain(self, spec, shape, name, **tags):
        with self.variable_scope:
            tags['trainable'] = tags.get('trainable', True)
            tags['regularizable'] = tags.get('regularizable', True)
            param = create_param(spec, shape, name, **tags)
            self.params[param] = set(
                tag for tag, value in list(tags.items()) if value)
            return param

    def add_param(self, spec, shape, name, **kwargs):
        param = self.add_param_plain(spec, shape, name, **kwargs)
        if name is not None and name.startswith(
                'W') and self.weight_normalization:
            # Hacky: check if the parameter is a weight matrix. If so, apply
            # weight normalization
            if len(param.get_shape()) == 2:
                v = param
                g = self.add_param_plain(
                    tf.ones_initializer(), (shape[1], ), name=name + '_wn/g')
                param = v * (tf.reshape(g, (1, -1)) / tf.sqrt(
                    tf.reduce_sum(tf.square(v), 0, keep_dims=True)))
            elif len(param.get_shape()) == 4:
                v = param
                g = self.add_param_plain(
                    tf.ones_initializer(), (shape[3], ), name=name + '_wn/g')
                param = v * (tf.reshape(g, (1, 1, 1, -1)) / tf.sqrt(
                    tf.reduce_sum(tf.square(v), [0, 1, 2], keep_dims=True)))
            else:
                raise NotImplementedError
        return param

    def get_params(self, **tags):
        result = list(self.params.keys())

        only = set(tag for tag, value in list(tags.items()) if value)
        if only:
            # retain all parameters that have all of the tags in `only`
            result = [
                param for param in result if not (only - self.params[param])
            ]

        exclude = set(tag for tag, value in list(tags.items()) if not value)
        if exclude:
            # retain all parameters that have none of the tags in `exclude`
            result = [
                param for param in result if not (self.params[param] & exclude)
            ]

        return result


class InputLayer(Layer):
    def __init__(self, shape, input_var=None, **kwargs):
        super(InputLayer, self).__init__(shape, **kwargs)
        self.shape = shape
        if input_var is None:
            with tf.compat.v1.variable_scope(self.name):
                input_var = tf.compat.v1.placeholder(
                    tf.float32, shape=shape, name='input')
        self.input_var = input_var

    @Layer.output_shape.getter
    def output_shape(self):
        return self.shape


class MergeLayer(Layer):
    def __init__(self, incomings, name=None, **kwargs):
        self.input_shapes = [
            incoming if isinstance(incoming, tuple) else incoming.output_shape
            for incoming in incomings
        ]
        self.input_layers = [
            None if isinstance(incoming, tuple) else incoming
            for incoming in incomings
        ]
        if name is None:
            name = '%s_%d' % (type(self).__name__, G._n_layers)
            G._n_layers += 1
        self.name = name
        self.params = OrderedDict()
        self.get_output_kwargs = []

    @Layer.output_shape.getter
    def output_shape(self):
        shape = self.get_output_shape_for(self.input_shapes)
        if any(isinstance(s, (tf.Variable, tf.Tensor)) for s in shape):
            raise ValueError('%s returned a symbolic output shape from its '
                             'get_output_shape_for() method: %r. This is not '
                             'allowed; shapes must be tuples of integers for '
                             'fixed-size dimensions and Nones for variable '
                             'dimensions.' % (self.__class__.__name__, shape))
        return shape

    def get_output_shape_for(self, input_shapes):
        raise NotImplementedError

    def get_output_for(self, inputs, **kwargs):
        raise NotImplementedError


class ConcatLayer(MergeLayer):
    """
    Concatenates multiple inputs along the specified axis. Inputs should have
    the same shape except for the dimension specified in axis, which can have
    different sizes.
    Parameters
    -----------
    incomings : a list of :class:`Layer` instances or tuples
        The layers feeding into this layer, or expected input shapes
    axis : int
        Axis which inputs are joined over
    """

    def __init__(self, incomings, axis=1, **kwargs):
        super(ConcatLayer, self).__init__(incomings, **kwargs)
        self.axis = axis

    def get_output_shape_for(self, input_shapes):
        # Infer the output shape by grabbing, for each axis, the first
        # input size that is not `None` (if there is any)
        output_shape = [
            next((s for s in sizes if s is not None), None)
            for sizes in zip(*input_shapes)
        ]

        def match(shape1, shape2):
            return (len(shape1) == len(shape2) and all(
                i == self.axis or s1 is None or s2 is None or s1 == s2
                for i, (s1, s2) in enumerate(zip(shape1, shape2))))

        # Check for compatibility with inferred output shape
        if not all(match(shape, output_shape) for shape in input_shapes):
            raise ValueError('Mismatch: input shapes must be the same except '
                             'in the concatenation axis')
        # Infer output shape on concatenation axis and return
        sizes = [input_shape[self.axis] for input_shape in input_shapes]
        concat_size = None if any(s is None for s in sizes) else sum(sizes)
        output_shape[self.axis] = concat_size
        return tuple(output_shape)

    def get_output_for(self, inputs, **kwargs):
        with tf.name_scope(self.name, values=[inputs]):
            dtypes = [x.dtype.as_numpy_dtype for x in inputs]
            if len(set(dtypes)) > 1:
                # need to convert to common data type
                common_dtype = np.core.numerictypes.find_common_type([],
                                                                     dtypes)
                inputs = [tf.cast(x, common_dtype) for x in inputs]
            return tf.concat(axis=self.axis, values=inputs)


concat = ConcatLayer  # shortcut


class XavierUniformInitializer:
    def __call__(self, shape, dtype=tf.float32, name=None, *args, **kwargs):
        if len(shape) == 2:
            n_inputs, n_outputs = shape
        else:
            receptive_field_size = np.prod(shape[:2])
            n_inputs = shape[-2] * receptive_field_size
            n_outputs = shape[-1] * receptive_field_size
        init_range = math.sqrt(6.0 / (n_inputs + n_outputs))
        with tf.name_scope(name, 'XavierUniformInitializer'):
            return tf.random_uniform_initializer(-init_range, init_range)(
                shape, dtype=dtype)


class HeUniformInitializer:
    def __call__(self, shape, dtype=tf.float32, name=None, *args, **kwargs):
        if len(shape) == 2:
            n_inputs, _ = shape
        else:
            receptive_field_size = np.prod(shape[:2])
            n_inputs = shape[-2] * receptive_field_size
        init_range = math.sqrt(1.0 / n_inputs)
        with tf.name_scope(name, 'HeUniformInitializer'):
            return tf.random_uniform_initializer(
                -init_range, init_range, dtype=dtype)(shape)


def py_ortho_init(scale):
    def _init(shape):
        u, s, v = np.linalg.svd(np.random.uniform(size=shape))
        return np.cast['float32'](u * scale)

    return _init


class OrthogonalInitializer:
    def __init__(self, scale=1.1):
        self.scale = scale

    def __call__(self, shape, dtype=tf.float32, *args, **kwargs):
        result, = tf.py_func(py_ortho_init(self.scale), [shape], [tf.float32])
        result.set_shape(shape)
        return result


class ParamLayer(Layer):
    def __init__(self,
                 incoming,
                 num_units,
                 param=tf.zeros_initializer(),
                 trainable=True,
                 **kwargs):
        super(ParamLayer, self).__init__(incoming, **kwargs)
        self.num_units = num_units
        self.param = self.add_param(
            param, (num_units, ), name='param', trainable=trainable)

    def get_output_shape_for(self, input_shape):
        return input_shape[:-1] + (self.num_units, )

    def get_output_for(self, input, **kwargs):
        with tf.name_scope(self.name, values=[input]):
            ndim = input.get_shape().ndims
            reshaped_param = tf.reshape(
                self.param, (1, ) * (ndim - 1) + (self.num_units, ))
            tile_arg = tf.concat(
                axis=0, values=[tf.shape(input)[:ndim - 1], [1]])
            tiled = tf.tile(reshaped_param, tile_arg)
            return tiled


class OpLayer(MergeLayer):
    def __init__(self,
                 incoming,
                 op,
                 shape_op=lambda x: x,
                 extras=None,
                 **kwargs):
        if extras is None:
            extras = []
        incomings = [incoming] + extras
        super(OpLayer, self).__init__(incomings, **kwargs)
        self.op = op
        self.shape_op = shape_op
        self.incomings = incomings

    def get_output_shape_for(self, input_shapes):
        return self.shape_op(*input_shapes)

    def get_output_for(self, inputs, **kwargs):
        with tf.name_scope(self.name, values=[inputs]):
            return self.op(*inputs)


class DenseLayer(Layer):
    def __init__(self,
                 incoming,
                 num_units,
                 nonlinearity=None,
                 w=XavierUniformInitializer(),
                 b=tf.zeros_initializer(),
                 **kwargs):
        super(DenseLayer, self).__init__(incoming, **kwargs)
        if nonlinearity is None:
            self.nonlinearity = tf.identity
        else:
            self.nonlinearity = nonlinearity

        self.num_units = num_units

        num_inputs = int(np.prod(self.input_shape[1:]))

        self.w = self.add_param(w, (num_inputs, num_units), name='W')
        if b is None:
            self.b = None
        else:
            self.b = self.add_param(
                b, (num_units, ), name='b', regularizable=False)

    def get_output_shape_for(self, input_shape):
        return (input_shape[0], self.num_units)

    def get_output_for(self, input, **kwargs):
        with tf.name_scope(self.name, values=[input]):
            if input.get_shape().ndims > 2:
                # if the input has more than two dimensions, flatten it into a
                # batch of feature vectors.
                input = tf.reshape(input, tf.stack([tf.shape(input)[0], -1]))
            activation = tf.matmul(input, self.w)
            if self.b is not None:
                activation = activation + tf.expand_dims(self.b, 0)
            return self.nonlinearity(activation)


class BaseConvLayer(Layer):
    def __init__(self,
                 incoming,
                 num_filters,
                 filter_size,
                 stride=1,
                 pad='VALID',
                 untie_biases=False,
                 w=XavierUniformInitializer(),
                 b=tf.zeros_initializer(),
                 nonlinearity=tf.nn.relu,
                 n=None,
                 **kwargs):
        """
        Input is assumed to be of shape batch*height*width*channels
        """
        super(BaseConvLayer, self).__init__(incoming, **kwargs)
        if nonlinearity is None:
            self.nonlinearity = tf.identity
        else:
            self.nonlinearity = nonlinearity

        if n is None:
            n = len(self.input_shape) - 2
        elif n != len(self.input_shape) - 2:
            raise ValueError('Tried to create a %dD convolution layer with '
                             'input shape %r. Expected %d input dimensions '
                             '(batchsize, channels, %d spatial dimensions).' %
                             (n, self.input_shape, n + 2, n))
        self.n = n
        self.num_filters = num_filters
        self.filter_size = as_tuple(filter_size, n, int)
        self.stride = as_tuple(stride, n, int)
        self.untie_biases = untie_biases

        self.pad = pad

        if pad == 'SAME':
            if any(s % 2 == 0 for s in self.filter_size):
                raise NotImplementedError(
                    '`same` padding requires odd filter size.')

        self.w = self.add_param(w, self.get_w_shape(), name='W')
        if b is None:
            self.b = None
        else:
            if self.untie_biases:
                biases_shape = self.output_shape[1:3] + (
                    num_filters, )  # + self.output_shape[2:]
            else:
                biases_shape = (num_filters, )
            self.b = self.add_param(
                b, biases_shape, name='b', regularizable=False)

    def get_w_shape(self):
        """Get the shape of the weight matrix `W`.
        Returns
        -------
        tuple of int
            The shape of the weight matrix.
        """
        num_input_channels = self.input_shape[-1]
        return self.filter_size + (num_input_channels, self.num_filters)

    def get_output_shape_for(self, input_shape):
        if self.pad == 'SAME':
            pad = ('same', ) * self.n
        elif self.pad == 'VALID':
            pad = (0, ) * self.n
        else:
            raise NotImplementedError

        # if isinstance(self.pad, tuple):
        #     pad = self.pad
        # else:
        #     pad = (self.pad,) * self.n
        batchsize = input_shape[0]
        return ((batchsize, ) + tuple(
            conv_output_length(input, filter, stride, p)
            for input, filter, stride, p in zip(
                input_shape[1:3], self.filter_size, self.stride, pad))) + (
                    self.num_filters, )

    def get_output_for(self, input, **kwargs):
        with tf.name_scope(self.name, values=[input]):
            conved = self.convolve(input, **kwargs)

            if self.b is None:
                activation = conved
            elif self.untie_biases:
                # raise NotImplementedError
                activation = conved + tf.expand_dims(self.b, 0)
            else:
                activation = conved + tf.reshape(self.b,
                                                 (1, 1, 1, self.num_filters))

            return self.nonlinearity(activation)

    def convolve(self, input, name=None, **kwargs):
        """
        Symbolically convolves `input` with ``self.w``, producing an output of
        shape ``self.output_shape``. To be implemented by subclasses.
        Parameters
        ----------
        input : TensorFlow tensor
            The input minibatch to convolve
        **kwargs
            Any additional keyword arguments from :meth:`get_output_for`
        Returns
        -------
        TensorFlow tensor
            `input` convolved according to the configuration of this layer,
            without any bias or nonlinearity applied.
        """
        raise NotImplementedError('BaseConvLayer does not implement the '
                                  'convolve() method. You will want to '
                                  'use a subclass such as Conv2DLayer.')


class Conv2DLayer(BaseConvLayer):
    def __init__(self,
                 incoming,
                 num_filters,
                 filter_size,
                 stride=(1, 1),
                 pad='VALID',
                 untie_biases=False,
                 w=XavierUniformInitializer(),
                 b=tf.zeros_initializer(),
                 nonlinearity=tf.nn.relu,
                 convolution=tf.nn.conv2d,
                 **kwargs):
        super(Conv2DLayer, self).__init__(
            incoming=incoming,
            num_filters=num_filters,
            filter_size=filter_size,
            stride=stride,
            pad=pad,
            untie_biases=untie_biases,
            w=w,
            b=b,
            nonlinearity=nonlinearity,
            n=2,
            **kwargs)
        self.convolution = convolution

    def convolve(self, input, name=None, **kwargs):
        with tf.name_scope(name, 'convolve', values=[input]):
            conved = self.convolution(
                input,
                self.w,
                strides=(1, ) + self.stride + (1, ),
                padding=self.pad)
            return conved


def pool_output_length(input_length, pool_size, stride, pad):
    if input_length is None or pool_size is None:
        return None

    if pad == 'SAME':
        return int(np.ceil(float(input_length) / float(stride)))

    return int(np.ceil(float(input_length - pool_size + 1) / float(stride)))


class Pool2DLayer(Layer):
    def __init__(self,
                 incoming,
                 pool_size,
                 stride=None,
                 pad='VALID',
                 mode='max',
                 **kwargs):
        super(Pool2DLayer, self).__init__(incoming, **kwargs)

        self.pool_size = as_tuple(pool_size, 2)

        if len(self.input_shape) != 4:
            raise ValueError('Tried to create a 2D pooling layer with '
                             'input shape %r. Expected 4 input dimensions '
                             '(batchsize, 2 spatial dimensions, channels).' %
                             (self.input_shape, ))

        if stride is None:
            self.stride = self.pool_size
        else:
            self.stride = as_tuple(stride, 2)

        self.pad = pad

        self.mode = mode

    def get_output_shape_for(self, input_shape):
        output_shape = list(input_shape)  # copy / convert to mutable list

        output_shape[1] = pool_output_length(
            input_shape[1],
            pool_size=self.pool_size[0],
            stride=self.stride[0],
            pad=self.pad,
        )

        output_shape[2] = pool_output_length(
            input_shape[2],
            pool_size=self.pool_size[1],
            stride=self.stride[1],
            pad=self.pad,
        )

        return tuple(output_shape)

    def get_output_for(self, input, **kwargs):
        with tf.name_scope(self.name, values=[input]):
            assert self.mode == 'max'
            pooled = tf.nn.max_pool(
                input,
                ksize=(1, ) + self.pool_size + (1, ),
                strides=(1, ) + self.stride + (1, ),
                padding=self.pad,
            )
            return pooled


def spatial_expected_softmax(x, temp=1):
    assert len(x.get_shape()) == 4
    vals = []
    for dim in [0, 1]:
        dim_val = x.get_shape()[dim + 1].value
        lin = tf.linspace(-1.0, 1.0, dim_val)
        lin = tf.expand_dims(lin, 1 - dim)
        lin = tf.expand_dims(lin, 0)
        lin = tf.expand_dims(lin, 3)
        m = tf.reduce_max(x, [1, 2], keep_dims=True)
        e = tf.exp((x - m) / temp) + 1e-5
        val = tf.reduce_sum(e * lin, [1, 2]) / (tf.reduce_sum(e, [1, 2]))
        vals.append(tf.expand_dims(val, 2))

    return tf.reshape(
        tf.concat(axis=2, values=vals), [-1, x.get_shape()[-1].value * 2])


class SpatialExpectedSoftmaxLayer(Layer):
    """
    Computes the softmax across a spatial region, separately for each channel,
    followed by an expectation operation.
    """

    def __init__(self, incoming, **kwargs):
        super().__init__(incoming, **kwargs)
        # self.temp = self.add_param(
        #     tf.ones_initializer, shape=(), name="temperature")

    def get_output_shape_for(self, input_shape):
        return (input_shape[0], input_shape[-1] * 2)

    def get_output_for(self, input, **kwargs):
        with tf.name_scope(self.name, values=[input]):
            return spatial_expected_softmax(input)  # , self.temp)
        # max_ = tf.reduce_max(input, reduction_indices=[1, 2], keep_dims=True)
        # exp = tf.exp(input - max_) + 1e-5

        # vals = []
        #
        # for dim in [0, 1]:
        #     dim_val = input.get_shape()[dim + 1].value
        #     lin = tf.linspace(-1.0, 1.0, dim_val)
        #     lin = tf.expand_dims(lin, 1 - dim)
        #     lin = tf.expand_dims(lin, 0)
        #     lin = tf.expand_dims(lin, 3)
        #     m = tf.reduce_max(input, [1, 2], keep_dims=True)
        #     e = tf.exp(input - m) + 1e-5
        #     val = tf.reduce_sum(e * lin, [1, 2]) / (tf.reduce_sum(e, [1, 2]))
        #     vals.append(tf.expand_dims(val, 2))
        #
        # return tf.reshape(
        #     tf.concat(2, vals), [-1, input.get_shape()[-1].value * 2])

        # import ipdb; ipdb.set_trace()

        # input.get_shape()
        # exp / tf.reduce_sum(exp, reduction_indices=[1, 2], keep_dims=True)
        # import ipdb;
        # ipdb.set_trace()
        # spatial softmax?

        # for dim in range(2):
        #     val = obs.get_shape()[dim + 1].value
        #     lin = tf.linspace(-1.0, 1.0, val)
        #     lin = tf.expand_dims(lin, 1 - dim)
        #     lin = tf.expand_dims(lin, 0)
        #     lin = tf.expand_dims(lin, 3)
        #     m = tf.reduce_max(e, [1, 2], keep_dims=True)
        #     e = tf.exp(e - m) + 1e-3
        #     val = tf.reduce_sum(e * lin, [1, 2]) / (tf.reduce_sum(e, [1, 2]))


class DropoutLayer(Layer):
    def __init__(self, incoming, p=0.5, rescale=True, **kwargs):
        super(DropoutLayer, self).__init__(incoming, **kwargs)
        self.p = p
        self.rescale = rescale

    def get_output_for(self, input, deterministic=False, **kwargs):
        """
        Parameters
        ----------
        input : tensor
            output from the previous layer
        deterministic : bool
            If true dropout and scaling is disabled, see notes
        """
        with tf.name_scope(self.name, values=[input]):
            if deterministic or self.p == 0:
                return input
            else:
                retain_prob = 1. - self.p
                if self.rescale:
                    input /= retain_prob

                # use nonsymbolic shape for dropout mask if possible
                return tf.nn.dropout(input, keep_prob=retain_prob)

    def get_output_shape_for(self, input_shape):
        return input_shape


# TODO: add Conv3DLayer


class FlattenLayer(Layer):
    """
    A layer that flattens its input. The leading ``outdim-1`` dimensions of
    the output will have the same shape as the input. The remaining dimensions
    are collapsed into the last dimension.
    Parameters
    ----------
    incoming : a :class:`Layer` instance or a tuple
        The layer feeding into this layer, or the expected input shape.
    outdim : int
        The number of dimensions in the output.
    See Also
    --------
    flatten  : Shortcut
    """

    def __init__(self, incoming, outdim=2, **kwargs):
        super(FlattenLayer, self).__init__(incoming, **kwargs)
        self.outdim = outdim

        if outdim < 1:
            raise ValueError('Dim must be >0, was %i', outdim)

    def get_output_shape_for(self, input_shape):
        to_flatten = input_shape[self.outdim - 1:]

        if any(s is None for s in to_flatten):
            flattened = None
        else:
            flattened = int(np.prod(to_flatten))

        return input_shape[:self.outdim - 1] + (flattened, )

    def get_output_for(self, input, **kwargs):
        with tf.name_scope(self.name, values=[input]):
            # total_entries = tf.reduce_prod(tf.shape(input))
            pre_shape = tf.shape(input)[:self.outdim - 1]
            to_flatten = tf.reduce_prod(tf.shape(input)[self.outdim - 1:])
            return tf.reshape(
                input,
                tf.concat(axis=0, values=[pre_shape,
                                          tf.stack([to_flatten])]))


flatten = FlattenLayer  # shortcut


class ReshapeLayer(Layer):
    def __init__(self, incoming, shape, **kwargs):
        super(ReshapeLayer, self).__init__(incoming, **kwargs)
        shape = tuple(shape)
        for s in shape:
            if isinstance(s, int):
                if s == 0 or s < -1:
                    raise ValueError('`shape` integers must be positive or -1')
            elif isinstance(s, list):
                if len(s) != 1 or not isinstance(s[0], int) or s[0] < 0:
                    raise ValueError('`shape` input references must be '
                                     'single-element lists of int >= 0')
            elif isinstance(s, (tf.Tensor, tf.Variable)):
                raise NotImplementedError
            else:
                raise ValueError('`shape` must be a tuple of int and/or [int]')
        if sum(s == -1 for s in shape) > 1:
            raise ValueError('`shape` cannot contain multiple -1')
        self.shape = shape
        # try computing the output shape once as a sanity check
        self.get_output_shape_for(self.input_shape)

    def get_output_shape_for(self, input_shape, **kwargs):
        # Initialize output shape from shape specification
        output_shape = list(self.shape)
        # First, replace all `[i]` with the corresponding input dimension, and
        # mask parts of the shapes thus becoming irrelevant for -1 inference
        masked_input_shape = list(input_shape)
        masked_output_shape = list(output_shape)
        for dim, o in enumerate(output_shape):
            if isinstance(o, list):
                if o[0] >= len(input_shape):
                    raise ValueError('specification contains [%d], but input '
                                     'shape has %d dimensions only' %
                                     (o[0], len(input_shape)))
                output_shape[dim] = input_shape[o[0]]
                masked_output_shape[dim] = input_shape[o[0]]
                if (input_shape[o[0]] is None) \
                        and (masked_input_shape[o[0]] is None):
                    # first time we copied this unknown input size: mask
                    # it, we have a 1:1 correspondence between out[dim] and
                    # in[o[0]] and can ignore it for -1 inference even if
                    # it is unknown.
                    masked_input_shape[o[0]] = 1
                    masked_output_shape[dim] = 1
        # Secondly, replace all symbolic shapes with `None`, as we cannot
        # infer their size here.
        for dim, o in enumerate(output_shape):
            if isinstance(o, (tf.Tensor, tf.Variable)):
                raise NotImplementedError
                # output_shape[dim] = None
                # masked_output_shape[dim] = None
        # From the shapes, compute the sizes of the input and output tensor
        input_size = (None if any(
            x is None
            for x in masked_input_shape) else np.prod(masked_input_shape))
        output_size = (None if any(
            x is None
            for x in masked_output_shape) else np.prod(masked_output_shape))
        del masked_input_shape, masked_output_shape
        # Finally, infer value for -1 if needed
        if -1 in output_shape:
            dim = output_shape.index(-1)
            if (input_size is None) or (output_size is None):
                output_shape[dim] = None
                output_size = None
            else:
                output_size *= -1
                output_shape[dim] = input_size // output_size
                output_size *= output_shape[dim]
        # Sanity check
        if (input_size is not None) and (output_size is not None) \
                and (input_size != output_size):
            raise ValueError(
                '%s cannot be reshaped to specification %s. '
                'The total size mismatches.' % (input_shape, self.shape))
        return tuple(output_shape)

    def get_output_for(self, input, **kwargs):
        with tf.name_scope(self.name, values=[input]):
            # Replace all `[i]` with the corresponding input dimension
            output_shape = list(self.shape)
            for dim, o in enumerate(output_shape):
                if isinstance(o, list):
                    output_shape[dim] = tf.shape(input)[o[0]]
            # Everything else is handled by TensorFlow
            return tf.reshape(input, tf.stack(output_shape))


reshape = ReshapeLayer  # shortcut


class SliceLayer(Layer):
    def __init__(self, incoming, indices, axis=-1, **kwargs):
        super(SliceLayer, self).__init__(incoming, **kwargs)
        self.slice = indices
        self.axis = axis

    def get_output_shape_for(self, input_shape):
        output_shape = list(input_shape)
        if isinstance(self.slice, int):
            del output_shape[self.axis]
        elif input_shape[self.axis] is not None:
            output_shape[self.axis] = len(
                list(range(*self.slice.indices(input_shape[self.axis]))))
        else:
            output_shape[self.axis] = None
        return tuple(output_shape)

    def get_output_for(self, input, **kwargs):
        with tf.name_scope(self.name, values=[input]):
            axis = self.axis
            ndims = input.get_shape().ndims
            if axis < 0:
                axis += ndims
            if isinstance(self.slice, int) and self.slice < 0:
                return tf.reverse(
                    input, [self.axis + 1
                            ])[(slice(None), ) * axis + (-1 - self.slice, ) +
                               (slice(None), ) * (ndims - axis - 1)]
            # import ipdb; ipdb.set_trace()
            return input[(slice(None), ) * axis + (self.slice, ) +
                         (slice(None), ) * (ndims - axis - 1)]


class DimshuffleLayer(Layer):
    def __init__(self, incoming, pattern, **kwargs):
        super(DimshuffleLayer, self).__init__(incoming, **kwargs)

        # Sanity check the pattern
        used_dims = set()
        for p in pattern:
            if isinstance(p, int):
                # Dimension p
                if p in used_dims:
                    raise ValueError('pattern contains dimension {0} more '
                                     'than once'.format(p))
                used_dims.add(p)
            elif p == 'x':
                # Broadcast
                pass
            else:
                raise ValueError('pattern should only contain dimension'
                                 "indices or 'x', not {0}".format(p))

        self.pattern = pattern

        # try computing the output shape once as a sanity check
        self.get_output_shape_for(self.input_shape)

    def get_output_shape_for(self, input_shape):
        # Build output shape while keeping track of the dimensions that we are
        # attempting to collapse, so we can ensure that they are broadcastable
        output_shape = []
        dims_used = [False] * len(input_shape)
        for p in self.pattern:
            if isinstance(p, int):
                if p < 0 or p >= len(input_shape):
                    raise ValueError('pattern contains {0}, but input shape '
                                     'has {1} dimensions '
                                     'only'.format(p, len(input_shape)))
                # Dimension p
                o = input_shape[p]
                dims_used[p] = True
            elif p == 'x':
                # Broadcast; will be of size 1
                o = 1
            output_shape.append(o)

        for i, (dim_size, used) in enumerate(zip(input_shape, dims_used)):
            if not used and dim_size != 1 and dim_size is not None:
                raise ValueError(
                    'pattern attempted to collapse dimension '
                    '{0} of size {1}; dimensions with size != 1/None are not'
                    'broadcastable and cannot be '
                    'collapsed'.format(i, dim_size))

        return tuple(output_shape)

    def get_output_for(self, input, **kwargs):
        with tf.name_scope(self.name, values=[input]):
            return tf.transpose(input, self.pattern)


dimshuffle = DimshuffleLayer  # shortcut


def apply_ln(layer):
    with tf.name_scope('apply_ln'):

        def _normalize(x, prefix):
            eps = 1e-5
            dim = x.get_shape()[-1].value

            bias_name = prefix + '_ln/bias'
            scale_name = prefix + '_ln/scale'

            if bias_name not in layer.norm_params:
                layer.norm_params[bias_name] = layer.add_param(
                    tf.zeros_initializer(), (dim, ),
                    name=bias_name,
                    regularizable=False)
            if scale_name not in layer.norm_params:
                layer.norm_params[scale_name] = layer.add_param(
                    tf.ones_initializer(), (dim, ), name=scale_name)

            bias = layer.norm_params[bias_name]
            scale = layer.norm_params[scale_name]
            mean, var = tf.nn.moments(x, axes=[1], keep_dims=True)
            x_normed = (x - mean) / tf.sqrt(var + eps)
            return x_normed * scale + bias

        return _normalize


class GRULayer(Layer):
    """
    A gated recurrent unit implements the following update mechanism:
    Reset gate:        r(t) = f_r(x(t) @ W_xr + h(t-1) @ W_hr + b_r)
    Update gate:       u(t) = f_u(x(t) @ W_xu + h(t-1) @ W_hu + b_u)
    Cell gate:         c(t) = f_c(x(t) @ W_xc + r(t) * (h(t-1) @ W_hc) + b_c)
    New hidden state:  h(t) = (1 - u(t)) * c(t) + u_t * h(t-1)
    Note that the reset, update, and cell vectors must have the same dimension
    as the hidden state
    """

    def __init__(self,
                 incoming,
                 num_units,
                 hidden_nonlinearity,
                 gate_nonlinearity=tf.nn.sigmoid,
                 w_x_init=XavierUniformInitializer(),
                 w_h_init=OrthogonalInitializer(),
                 b_init=tf.zeros_initializer(),
                 hidden_init=tf.zeros_initializer(),
                 hidden_init_trainable=False,
                 layer_normalization=False,
                 **kwargs):

        if hidden_nonlinearity is None:
            hidden_nonlinearity = tf.identity

        if gate_nonlinearity is None:
            gate_nonlinearity = tf.identity

        super(GRULayer, self).__init__(incoming, **kwargs)

        with tf.compat.v1.variable_scope('gru_layer_step'):
            input_shape = self.input_shape[2:]

            input_dim = np.prod(input_shape)

            self.layer_normalization = layer_normalization

            # Weights for the initial hidden state
            self.h0 = self.add_param(
                hidden_init, (num_units, ),
                name='h0',
                trainable=hidden_init_trainable,
                regularizable=False)
            # Weights for the reset gate
            self.W_xr = self.add_param(
                w_x_init, (input_dim, num_units), name='W_xr')
            self.W_hr = self.add_param(
                w_h_init, (num_units, num_units), name='W_hr')
            self.b_r = self.add_param(
                b_init, (num_units, ), name='b_r', regularizable=False)
            # Weights for the update gate
            self.W_xu = self.add_param(
                w_x_init, (input_dim, num_units), name='W_xu')
            self.W_hu = self.add_param(
                w_h_init, (num_units, num_units), name='W_hu')
            self.b_u = self.add_param(
                b_init, (num_units, ), name='b_u', regularizable=False)
            # Weights for the cell gate
            self.W_xc = self.add_param(
                w_x_init, (input_dim, num_units), name='W_xc')
            self.W_hc = self.add_param(
                w_h_init, (num_units, num_units), name='W_hc')
            self.b_c = self.add_param(
                b_init, (num_units, ), name='b_c', regularizable=False)

            self.W_x_ruc = tf.concat(
                axis=1,
                values=[self.W_xr, self.W_xu, self.W_xc],
                name='W_x_ruc')
            self.W_h_ruc = tf.concat(
                axis=1,
                values=[self.W_hr, self.W_hu, self.W_hc],
                name='W_h_ruc')
            self.W_x_ru = tf.concat(
                axis=1, values=[self.W_xr, self.W_xu], name='W_x_ru')
            self.W_h_ru = tf.concat(
                axis=1, values=[self.W_hr, self.W_hu], name='W_h_ru')
            self.b_ruc = tf.concat(
                axis=0, values=[self.b_r, self.b_u, self.b_c], name='b_ruc')

            self.gate_nonlinearity = gate_nonlinearity
            self.num_units = num_units
            self.nonlinearity = hidden_nonlinearity
            self.norm_params = dict()

            # pre-run the step method to initialize the normalization
            # parameters
            h_dummy = tf.compat.v1.placeholder(
                dtype=tf.float32, shape=(None, num_units), name='h_dummy')
            x_dummy = tf.compat.v1.placeholder(
                dtype=tf.float32, shape=(None, input_dim), name='x_dummy')
            self.step(h_dummy, x_dummy)

    def step(self, hprev, x, name=None):
        with tf.name_scope(name, 'step', values=[hprev, x]):
            if self.layer_normalization:
                ln = apply_ln(self)
                x_ru = ln(tf.matmul(x, self.W_x_ru), 'x_ru')
                h_ru = ln(tf.matmul(hprev, self.W_h_ru), 'h_ru')
                x_r, x_u = tf.split(axis=1, num_or_size_splits=2, value=x_ru)
                h_r, h_u = tf.split(axis=1, num_or_size_splits=2, value=h_ru)
                x_c = ln(tf.matmul(x, self.W_xc), 'x_c')
                h_c = ln(tf.matmul(hprev, self.W_hc), 'h_c')
                r = self.gate_nonlinearity(x_r + h_r)
                u = self.gate_nonlinearity(x_u + h_u)
                c = self.nonlinearity(x_c + r * h_c)
                h = (1 - u) * c + u * hprev
                return h
            else:
                xb_ruc = tf.matmul(x, self.W_x_ruc) + tf.reshape(
                    self.b_ruc, (1, -1))
                h_ruc = tf.matmul(hprev, self.W_h_ruc)
                xb_r, xb_u, xb_c = tf.split(
                    axis=1, num_or_size_splits=3, value=xb_ruc)
                h_r, h_u, h_c = tf.split(
                    axis=1, num_or_size_splits=3, value=h_ruc)
                r = self.gate_nonlinearity(xb_r + h_r)
                u = self.gate_nonlinearity(xb_u + h_u)
                c = self.nonlinearity(xb_c + r * h_c)
                h = (1 - u) * c + u * hprev
                return h

    def get_step_layer(self, l_in, l_prev_hidden, name=None):
        return GRUStepLayer(
            incomings=[l_in, l_prev_hidden], recurrent_layer=self, name=name)

    def get_output_shape_for(self, input_shape):
        n_batch, n_steps = input_shape[:2]
        return n_batch, n_steps, self.num_units

    def get_output_for(self, input, **kwargs):
        with tf.name_scope(self.name, values=[input]):
            input_shape = tf.shape(input)
            n_batches = input_shape[0]
            n_steps = input_shape[1]
            input = tf.reshape(input, tf.stack([n_batches, n_steps, -1]))
            if ('recurrent_state' in kwargs
                    and self in kwargs['recurrent_state']):
                h0s = kwargs['recurrent_state'][self]
            else:
                h0s = tf.tile(
                    tf.reshape(self.h0, (1, self.num_units)), (n_batches, 1))
            # flatten extra dimensions
            shuffled_input = tf.transpose(input, (1, 0, 2))
            hs = tf.scan(self.step, elems=shuffled_input, initializer=h0s)
            shuffled_hs = tf.transpose(hs, (1, 0, 2))
            if 'recurrent_state_output' in kwargs:
                kwargs['recurrent_state_output'][self] = shuffled_hs
            return shuffled_hs


class GRUStepLayer(MergeLayer):
    def __init__(self, incomings, recurrent_layer, **kwargs):
        super(GRUStepLayer, self).__init__(incomings, **kwargs)
        self._gru_layer = recurrent_layer

    def get_params(self, **tags):
        return self._gru_layer.get_params(**tags)

    def get_output_shape_for(self, input_shapes):
        n_batch = input_shapes[0][0]
        return n_batch, self._gru_layer.num_units

    def get_output_for(self, inputs, **kwargs):
        with tf.name_scope(self.name, values=[inputs]):
            x, hprev = inputs
            n_batch = tf.shape(x)[0]
            x = tf.reshape(x, tf.stack([n_batch, -1]))
            x.set_shape((None, self.input_shapes[0][1]))
            return self._gru_layer.step(hprev, x)


class TfGRULayer(Layer):
    """
    Use TensorFlow's built-in GRU implementation
    """

    def __init__(self,
                 incoming,
                 num_units,
                 hidden_nonlinearity,
                 horizon=None,
                 hidden_init_trainable=False,
                 **kwargs):
        assert len(incoming.output_shape) == 3
        input_dim = incoming.shape[2]
        gru = tf.nn.rnn_cell.GRUCell(
            num_units=num_units, activation=hidden_nonlinearity)
        self.num_units = num_units
        self.horizon = horizon
        self.gru = gru
        self.hidden_nonlinearity = hidden_nonlinearity
        Layer.__init__(self, incoming=incoming, **kwargs)
        # dummy input variable

        with tf.compat.v1.variable_scope(self.name) as vs:
            input_dummy = tf.compat.v1.placeholder(
                tf.float32, (None, input_dim), 'input_dummy')
            hidden_dummy = tf.compat.v1.placeholder(
                tf.float32, (None, num_units), 'hidden_dummy')
            gru(input_dummy, hidden_dummy, scope=vs)
            vs.reuse_variables()
            self.scope = vs
            all_vars = [
                v for v in tf.compat.v1.global_variables()
                if v.name.startswith(vs.name)
            ]
            trainable_vars = [
                v for v in tf.compat.v1.trainable_variables()
                if v.name.startswith(vs.name)
            ]

        for var in trainable_vars:
            self.add_param(spec=var, shape=None, name=None, trainable=True)
        for var in set(all_vars) - set(trainable_vars):
            self.add_param(spec=var, shape=None, name=None, trainable=False)
        self.h0 = self.add_param(
            tf.zeros_initializer(), (num_units, ),
            name='h0',
            trainable=hidden_init_trainable,
            regularizable=False)

    def step(self, hprev, x, name=None):
        with tf.name_scope(name, 'step', values=[hprev, x]):
            return self.gru(x, hprev, scope=self.scope)[1]

    def get_output_for(self, input, **kwargs):
        with tf.name_scope(self.name, values=[input]):
            input_shape = tf.shape(input)
            n_batches = input_shape[0]
            state = tf.tile(
                tf.reshape(self.h0, (1, self.num_units)), (n_batches, 1))
            state.set_shape((None, self.num_units))
            if self.horizon is not None:
                outputs = []
                for idx in range(self.horizon):
                    output, state = self.gru(
                        input[:, idx, :], state,
                        scope=self.scope)  # self.name)
                    outputs.append(tf.expand_dims(output, 1))
                outputs = tf.concat(axis=1, values=outputs)
                return outputs
            else:
                n_steps = input_shape[1]
                input = tf.reshape(input, tf.stack([n_batches, n_steps, -1]))
                # flatten extra dimensions
                shuffled_input = tf.transpose(input, (1, 0, 2))
                shuffled_input.set_shape((None, None, self.input_shape[-1]))
                hs = tf.scan(
                    self.step, elems=shuffled_input, initializer=state)
                shuffled_hs = tf.transpose(hs, (1, 0, 2))
                return shuffled_hs

    def get_output_shape_for(self, input_shape):
        n_batch, n_steps = input_shape[:2]
        return n_batch, n_steps, self.num_units

    def get_step_layer(self, l_in, l_prev_hidden, name=None):
        return GRUStepLayer(
            incomings=[l_in, l_prev_hidden], recurrent_layer=self, name=name)


class PseudoLSTMLayer(Layer):
    """
    A Pseudo LSTM unit implements the following update mechanism:

    Incoming gate:     i(t) = σ(W_hi @ h(t-1)) + W_xi @ x(t) + b_i)
    Forget gate:       f(t) = σ(W_hf @ h(t-1)) + W_xf @ x(t) + b_f)
    Out gate:          o(t) = σ(W_ho @ h(t-1)) + W_xo @ x(t) + b_o)
    New cell gate:     c_new(t) = ϕ(W_hc @ (o(t) * h(t-1)) + W_xc @ x(t) + b_c)
    Cell gate:         c(t) = f(t) * c(t-1) + i(t) * c_new(t)
    Hidden state:      h(t) = ϕ(c(t))
    Output:            out  = h(t)

    If gate_squash_inputs is set to True, we have the following updates
    instead:

    Out gate:          o(t) = σ(W_ho @ h(t-1)) + W_xo @ x(t) + b_o)
    Incoming gate:     i(t) = σ(W_hi @ (o(t) * h(t-1)) + W_xi @ x(t) + b_i)
    Forget gate:       f(t) = σ(W_hf @ (o(t) * h(t-1)) + W_xf @ x(t) + b_f)
    New cell gate:     c_new(t) = ϕ(W_hc @ (o(t) * h(t-1)) + W_xc @ x(t) + b_c)
    Cell state:        c(t) = f(t) * c(t-1) + i(t) * c_new(t)
    Hidden state:      h(t) = ϕ(c(t))
    Output:            out  = h(t)

    Note that the incoming, forget, cell, and out vectors must have the same
    dimension as the hidden state

    The notation is slightly different from
    http://r2rt.com/written-memories-understanding-deriving-and-extending-the-lstm.html.
    Here we introduce the cell gate and swap its role with the hidden state, so
    that the output is the same as the hidden state (and we can use this as a
    drop-in replacement for LSTMLayer).
    """

    def __init__(self,
                 incoming,
                 num_units,
                 hidden_nonlinearity=tf.tanh,
                 gate_nonlinearity=tf.nn.sigmoid,
                 w_x_init=XavierUniformInitializer(),
                 w_h_init=OrthogonalInitializer(),
                 forget_bias=1.0,
                 b_init=tf.zeros_initializer(),
                 hidden_init=tf.zeros_initializer(),
                 hidden_init_trainable=False,
                 cell_init=tf.zeros_initializer(),
                 cell_init_trainable=False,
                 gate_squash_inputs=False,
                 layer_normalization=False,
                 **kwargs):

        if hidden_nonlinearity is None:
            hidden_nonlinearity = tf.identity

        if gate_nonlinearity is None:
            gate_nonlinearity = tf.identity

        super(PseudoLSTMLayer, self).__init__(incoming, **kwargs)

        with tf.compat.v1.variable_scope('pseudo_lstm_layer'):
            self.layer_normalization = layer_normalization

            input_shape = self.input_shape[2:]

            input_dim = np.prod(input_shape)
            # Weights for the initial hidden state (this is actually not used,
            # since the initial hidden state is determined by the initial cell
            # state via h0 = self.nonlinearity(c0)). It is here merely for
            # interface convenience
            self.h0 = self.add_param(
                hidden_init, (num_units, ),
                name='h0',
                trainable=hidden_init_trainable,
                regularizable=False)
            # Weights for the initial cell state
            self.c0 = self.add_param(
                cell_init, (num_units, ),
                name='c0',
                trainable=cell_init_trainable,
                regularizable=False)
            # Weights for the incoming gate
            self.W_xi = self.add_param(
                w_x_init, (input_dim, num_units), name='W_xi')
            self.W_hi = self.add_param(
                w_h_init, (num_units, num_units), name='W_hi')
            self.b_i = self.add_param(
                b_init, (num_units, ), name='b_i', regularizable=False)
            # Weights for the forget gate
            self.W_xf = self.add_param(
                w_x_init, (input_dim, num_units), name='W_xf')
            self.W_hf = self.add_param(
                w_h_init, (num_units, num_units), name='W_hf')
            self.b_f = self.add_param(
                b_init, (num_units, ), name='b_f', regularizable=False)
            # Weights for the out gate
            self.W_xo = self.add_param(
                w_x_init, (input_dim, num_units), name='W_xo')
            self.W_ho = self.add_param(
                w_h_init, (num_units, num_units), name='W_ho')
            self.b_o = self.add_param(
                b_init, (num_units, ), name='b_o', regularizable=False)
            # Weights for the cell gate
            self.W_xc = self.add_param(
                w_x_init, (input_dim, num_units), name='W_xc')
            self.W_hc = self.add_param(
                w_h_init, (num_units, num_units), name='W_hc')
            self.b_c = self.add_param(
                b_init, (num_units, ), name='b_c', regularizable=False)

            self.gate_nonlinearity = gate_nonlinearity
            self.num_units = num_units
            self.nonlinearity = hidden_nonlinearity
            self.forget_bias = forget_bias
            self.gate_squash_inputs = gate_squash_inputs

            self.W_x_ifo = tf.concat(
                axis=1, values=[self.W_xi, self.W_xf, self.W_xo])
            self.W_h_ifo = tf.concat(
                axis=1, values=[self.W_hi, self.W_hf, self.W_ho])

            self.W_x_if = tf.concat(axis=1, values=[self.W_xi, self.W_xf])
            self.W_h_if = tf.concat(axis=1, values=[self.W_hi, self.W_hf])

            self.norm_params = dict()

    def step(self, hcprev, x, name=None):
        with tf.name_scope(name, 'step', values=[hcprev, x]):
            hprev = hcprev[:, :self.num_units]
            cprev = hcprev[:, self.num_units:]

            if self.layer_normalization:
                ln = apply_ln(self)
            else:

                def f(x, *args):
                    return x

                ln = f

            if self.gate_squash_inputs:
                """
                Out gate:          o(t) = σ(W_ho @ h(t-1)) + W_xo @ x(t) + b_o)
                Incoming gate:     i(t) = σ(W_hi @ (o(t) * h(t-1)) +
                                            W_xi @ x(t) + b_i)
                Forget gate:       f(t) = σ(W_hf @ (o(t) * h(t-1)) +
                                            W_xf @ x(t) + b_f)
                New cell gate:     c_new(t) = ϕ(W_hc @ (o(t) * h(t-1)) +
                                                W_xc @ x(t) + b_c)
                Cell state:        c(t) = f(t) * c(t-1) + i(t) * c_new(t)
                Hidden state:      h(t) = ϕ(c(t))
                Output:            out  = h(t)
                """

                o = self.nonlinearity(
                    ln(tf.matmul(hprev, self.W_ho), 'h_o') +
                    ln(tf.matmul(x, self.W_xo), 'x_o') + self.b_o)

                x_if = ln(tf.matmul(x, self.W_x_if), 'x_if')
                h_if = ln(tf.matmul(o * hprev, self.W_h_if), 'h_if')

                x_i, x_f = tf.split(axis=1, num_or_size_splits=2, value=x_if)
                h_i, h_f = tf.split(axis=1, num_or_size_splits=2, value=h_if)

                i = self.gate_nonlinearity(x_i + h_i + self.b_i)
                f = self.gate_nonlinearity(x_f + h_f + self.b_f +
                                           self.forget_bias)
                c_new = self.nonlinearity(
                    ln(tf.matmul(o * hprev, self.W_hc), 'h_c') +
                    ln(tf.matmul(x, self.W_xc), 'x_c') + self.b_c)
                c = f * cprev + i * c_new
                h = self.nonlinearity(ln(c, 'c'))
                return tf.concat(axis=1, values=[h, c])
            else:
                """
                Incoming gate:     i(t) = σ(W_hi @ h(t-1)) + W_xi @ x(t) + b_i)
                Forget gate:       f(t) = σ(W_hf @ h(t-1)) + W_xf @ x(t) + b_f)
                Out gate:          o(t) = σ(W_ho @ h(t-1)) + W_xo @ x(t) + b_o)
                New cell gate:     c_new(t) = ϕ(W_hc @ (o(t) * h(t-1)) +
                                                W_xc @ x(t) + b_c)
                Cell gate:         c(t) = f(t) * c(t-1) + i(t) * c_new(t)
                Hidden state:      h(t) = ϕ(c(t))
                Output:            out  = h(t)
                """

                x_ifo = ln(tf.matmul(x, self.W_x_ifo), 'x_ifo')
                h_ifo = ln(tf.matmul(hprev, self.W_h_ifo), 'h_ifo')

                x_i, x_f, x_o = tf.split(
                    axis=1, num_or_size_splits=3, value=x_ifo)
                h_i, h_f, h_o = tf.split(
                    axis=1, num_or_size_splits=3, value=h_ifo)

                i = self.gate_nonlinearity(x_i + h_i + self.b_i)
                f = self.gate_nonlinearity(x_f + h_f + self.b_f +
                                           self.forget_bias)
                o = self.gate_nonlinearity(x_o + h_o + self.b_o)
                c_new = self.nonlinearity(
                    ln(tf.matmul(o * hprev, self.W_hc), 'h_c') +
                    ln(tf.matmul(x, self.W_xc), 'x_c') + self.b_c)
                c = f * cprev + i * c_new
                h = self.nonlinearity(ln(c, 'c'))
                return tf.concat(axis=1, values=[h, c])

    def get_step_layer(self, l_in, l_prev_state, name=None):
        return LSTMStepLayer(
            incomings=[l_in, l_prev_state], recurrent_layer=self, name=name)

    def get_output_shape_for(self, input_shape):
        n_batch, n_steps = input_shape[:2]
        return n_batch, n_steps, self.num_units

    def get_output_for(self, input, **kwargs):
        with tf.name_scope(self.name, values=[input]):
            input_shape = tf.shape(input)
            n_batches = input_shape[0]
            n_steps = input_shape[1]
            input = tf.reshape(input, tf.stack([n_batches, n_steps, -1]))
            c0s = tf.tile(
                tf.reshape(self.c0, (1, self.num_units)), (n_batches, 1))
            h0s = self.nonlinearity(c0s)
            # flatten extra dimensions
            shuffled_input = tf.transpose(input, (1, 0, 2))
            hcs = tf.scan(
                self.step,
                elems=shuffled_input,
                initializer=tf.concat(axis=1, values=[h0s, c0s]))
            shuffled_hcs = tf.transpose(hcs, (1, 0, 2))
            shuffled_hs = shuffled_hcs[:, :, :self.num_units]
            return shuffled_hs


class LSTMLayer(Layer):
    """
    A LSTM unit implements the following update mechanism:

    Incoming gate:    i(t) = f_i(x(t) @ W_xi + h(t-1) @ W_hi +
                                 w_ci * c(t-1) + b_i)
    Forget gate:      f(t) = f_f(x(t) @ W_xf + h(t-1) @ W_hf +
                                 w_cf * c(t-1) + b_f)
    Cell gate:        c(t) = f(t) * c(t - 1) + i(t) * f_c(x(t) @ W_xc +
                             h(t-1) @ W_hc + b_c)
    Out gate:         o(t) = f_o(x(t) @ W_xo + h(t-1) W_ho + w_co * c(t) + b_o)
    New hidden state: h(t) = o(t) * f_h(c(t))

    Note that the incoming, forget, cell, and out vectors must have the same
    dimension as the hidden state
    """

    def __init__(self,
                 incoming,
                 num_units,
                 hidden_nonlinearity=tf.tanh,
                 gate_nonlinearity=tf.nn.sigmoid,
                 w_x_init=XavierUniformInitializer(),
                 w_h_init=OrthogonalInitializer(),
                 forget_bias=1.0,
                 use_peepholes=False,
                 w_init=tf.random_normal_initializer(stddev=0.1),
                 b_init=tf.zeros_initializer(),
                 hidden_init=tf.zeros_initializer(),
                 hidden_init_trainable=False,
                 cell_init=tf.zeros_initializer(),
                 cell_init_trainable=False,
                 layer_normalization=False,
                 **kwargs):

        if hidden_nonlinearity is None:
            hidden_nonlinearity = tf.identity

        if gate_nonlinearity is None:
            gate_nonlinearity = tf.identity

        super(LSTMLayer, self).__init__(incoming, **kwargs)

        with tf.compat.v1.variable_scope('lstm_layer'):
            self.layer_normalization = layer_normalization

            input_shape = self.input_shape[2:]

            input_dim = np.prod(input_shape)
            # Weights for the initial hidden state
            self.h0 = self.add_param(
                hidden_init, (num_units, ),
                name='h0',
                trainable=hidden_init_trainable,
                regularizable=False)
            # Weights for the initial cell state
            self.c0 = self.add_param(
                cell_init, (num_units, ),
                name='c0',
                trainable=cell_init_trainable,
                regularizable=False)
            # Weights for the incoming gate
            self.W_xi = self.add_param(
                w_x_init, (input_dim, num_units), name='W_xi')
            self.W_hi = self.add_param(
                w_h_init, (num_units, num_units), name='W_hi')
            if use_peepholes:
                self.w_ci = self.add_param(w_init, (num_units, ), name='w_ci')
            self.b_i = self.add_param(
                b_init, (num_units, ), name='b_i', regularizable=False)
            # Weights for the forget gate
            self.W_xf = self.add_param(
                w_x_init, (input_dim, num_units), name='W_xf')
            self.W_hf = self.add_param(
                w_h_init, (num_units, num_units), name='W_hf')
            if use_peepholes:
                self.w_cf = self.add_param(w_init, (num_units, ), name='w_cf')
            self.b_f = self.add_param(
                b_init, (num_units, ), name='b_f', regularizable=False)
            # Weights for the cell gate
            self.W_xc = self.add_param(
                w_x_init, (input_dim, num_units), name='W_xc')
            self.W_hc = self.add_param(
                w_h_init, (num_units, num_units), name='W_hc')
            self.b_c = self.add_param(
                b_init, (num_units, ), name='b_c', regularizable=False)
            # Weights for the reset gate
            self.W_xr = self.add_param(
                w_x_init, (input_dim, num_units), name='W_xr')
            self.W_hr = self.add_param(
                w_h_init, (num_units, num_units), name='W_hr')
            self.b_r = self.add_param(
                b_init, (num_units, ), name='b_r', regularizable=False)
            # Weights for the out gate
            self.W_xo = self.add_param(
                w_x_init, (input_dim, num_units), name='W_xo')
            self.W_ho = self.add_param(
                w_h_init, (num_units, num_units), name='W_ho')
            if use_peepholes:
                self.w_co = self.add_param(w_init, (num_units, ), name='w_co')
            self.b_o = self.add_param(
                b_init, (num_units, ), name='b_o', regularizable=False)
            self.gate_nonlinearity = gate_nonlinearity
            self.num_units = num_units
            self.nonlinearity = hidden_nonlinearity
            self.forget_bias = forget_bias
            self.use_peepholes = use_peepholes

            self.W_x_ifco = tf.concat(
                axis=1, values=[self.W_xi, self.W_xf, self.W_xc, self.W_xo])
            self.W_h_ifco = tf.concat(
                axis=1, values=[self.W_hi, self.W_hf, self.W_hc, self.W_ho])

            if use_peepholes:
                self.w_c_ifo = tf.concat(
                    axis=0, values=[self.w_ci, self.w_cf, self.w_co])

            self.norm_params = dict()

    def step(self, hcprev, x, name=None):
        """
        Incoming gate:     i(t) = f_i(x(t) @ W_xi + h(t-1) @ W_hi +
                                      w_ci * c(t-1) + b_i)
        Forget gate:       f(t) = f_f(x(t) @ W_xf + h(t-1) @ W_hf +
                                      w_cf * c(t-1) + b_f)
        Cell gate:         c(t) = f(t) * c(t - 1) + i(t) * f_c(x(t) @ W_xc +
                                  h(t-1) @ W_hc + b_c)
        Out gate:          o(t) = f_o(x(t) @ W_xo + h(t-1) W_ho +
                                      w_co * c(t) + b_o)
        New hidden state:  h(t) = o(t) * f_h(c(t))
        """
        with tf.name_scope(name, 'step', values=[hcprev, x]):
            hprev = hcprev[:, :self.num_units]
            cprev = hcprev[:, self.num_units:]

            if self.layer_normalization:
                ln = apply_ln(self)
            else:

                def f(x, *args):
                    return x

                ln = f

            x_ifco = ln(tf.matmul(x, self.W_x_ifco), 'x_ifco')
            h_ifco = ln(tf.matmul(hprev, self.W_h_ifco), 'h_ifco')
            x_i, x_f, x_c, x_o = tf.split(
                axis=1, num_or_size_splits=4, value=x_ifco)
            h_i, h_f, h_c, h_o = tf.split(
                axis=1, num_or_size_splits=4, value=h_ifco)

            if self.use_peepholes:
                i = self.gate_nonlinearity(x_i + h_i + self.w_ci * cprev +
                                           self.b_i)
                f = self.gate_nonlinearity(x_f + h_f + self.w_cf * cprev +
                                           self.b_f + self.forget_bias)

                o = self.gate_nonlinearity(x_o + h_o + self.w_co * cprev +
                                           self.b_o)
            else:
                i = self.gate_nonlinearity(x_i + h_i + self.b_i)
                f = self.gate_nonlinearity(x_f + h_f + self.b_f +
                                           self.forget_bias)
                o = self.gate_nonlinearity(x_o + h_o + self.b_o)

            c = f * cprev + i * self.nonlinearity(x_c + h_c + self.b_c)
            h = o * self.nonlinearity(ln(c, 'c'))

            return tf.concat(axis=1, values=[h, c])

    def get_step_layer(self, l_in, l_prev_state, name=None):
        return LSTMStepLayer(
            incomings=[l_in, l_prev_state], recurrent_layer=self, name=name)

    def get_output_shape_for(self, input_shape):
        n_batch, n_steps = input_shape[:2]
        return n_batch, n_steps, self.num_units

    def get_output_for(self, input, **kwargs):
        with tf.name_scope(self.name, values=[input]):
            input_shape = tf.shape(input)
            n_batches = input_shape[0]
            n_steps = input_shape[1]
            input = tf.reshape(input, tf.stack([n_batches, n_steps, -1]))
            h0s = tf.tile(
                tf.reshape(self.h0, (1, self.num_units)), (n_batches, 1))
            c0s = tf.tile(
                tf.reshape(self.c0, (1, self.num_units)), (n_batches, 1))
            # flatten extra dimensions
            shuffled_input = tf.transpose(input, (1, 0, 2))
            hcs = tf.scan(
                self.step,
                elems=shuffled_input,
                initializer=tf.concat(axis=1, values=[h0s, c0s]))
            shuffled_hcs = tf.transpose(hcs, (1, 0, 2))
            shuffled_hs = shuffled_hcs[:, :, :self.num_units]
            if 'recurrent_state_output' in kwargs:
                kwargs['recurrent_state_output'][self] = shuffled_hcs
            return shuffled_hs


class LSTMStepLayer(MergeLayer):
    def __init__(self, incomings, recurrent_layer, **kwargs):
        super(LSTMStepLayer, self).__init__(incomings, **kwargs)
        self._recurrent_layer = recurrent_layer

    def get_params(self, **tags):
        return self._recurrent_layer.get_params(**tags)

    def get_output_shape_for(self, input_shapes):
        n_batch = input_shapes[0][0]
        return n_batch, 2 * self._recurrent_layer.num_units

    def get_output_for(self, inputs, **kwargs):
        with tf.name_scope(self.name, values=[inputs]):
            x, hcprev = inputs
            n_batch = tf.shape(x)[0]
            x = tf.reshape(x, tf.stack([n_batch, -1]))
            hc = self._recurrent_layer.step(hcprev, x)
            return hc


class TfBasicLSTMLayer(Layer):
    """
    Use TensorFlow's built-in (basic) LSTM implementation
    """

    def __init__(self,
                 incoming,
                 num_units,
                 hidden_nonlinearity,
                 horizon=None,
                 hidden_init_trainable=False,
                 forget_bias=1.0,
                 use_peepholes=False,
                 **kwargs):
        assert not use_peepholes, 'Basic LSTM does not support peepholes!'
        assert len(incoming.output_shape) == 3
        input_dim = incoming.shape[2]
        lstm = tf.contrib.rnn.BasicLSTMCell(
            num_units=num_units,
            activation=hidden_nonlinearity,
            state_is_tuple=True,
            forget_bias=forget_bias)
        self.num_units = num_units
        self.horizon = horizon
        self.lstm = lstm
        self.hidden_nonlinearity = hidden_nonlinearity
        Layer.__init__(self, incoming=incoming, **kwargs)

        with tf.compat.v1.variable_scope(self.name):
            # dummy input variable
            input_dummy = tf.compat.v1.placeholder(
                tf.float32, (None, input_dim), 'input_dummy')
            hidden_dummy = tf.compat.v1.placeholder(
                tf.float32, (None, num_units), 'hidden_dummy')
            cell_dummy = tf.compat.v1.placeholder(
                tf.float32, (None, num_units), 'cell_dummy')

            with tf.compat.v1.variable_scope('lstm') as vs:
                lstm(input_dummy, (cell_dummy, hidden_dummy), scope=vs)
                vs.reuse_variables()
                self.scope = vs
                all_vars = [
                    v for v in tf.compat.v1.global_variables()
                    if v.name.startswith(vs.name)
                ]
                trainable_vars = [
                    v for v in tf.compat.v1.trainable_variables()
                    if v.name.startswith(vs.name)
                ]

        for var in trainable_vars:
            self.add_param(spec=var, shape=None, name=None, trainable=True)
        for var in set(all_vars) - set(trainable_vars):
            self.add_param(spec=var, shape=None, name=None, trainable=False)

        self.h0 = self.add_param(
            tf.zeros_initializer(), (num_units, ),
            name='h0',
            trainable=hidden_init_trainable,
            regularizable=False)
        self.c0 = self.add_param(
            tf.zeros_initializer(), (num_units, ),
            name='c0',
            trainable=hidden_init_trainable,
            regularizable=False)

    def step(self, hcprev, x, name=None):
        with tf.name_scope(name, 'step', values=[hcprev, x]):
            hprev = hcprev[:, :self.num_units]
            cprev = hcprev[:, self.num_units:]
            x.set_shape((None, self.input_shape[-1]))
            c, h = self.lstm(x, (cprev, hprev), scope=self.scope)[1]
            return tf.concat(axis=1, values=[h, c])

    def get_output_for(self, input, **kwargs):
        with tf.name_scope(self.name, values=[input]):
            input_shape = tf.shape(input)
            n_batches = input_shape[0]
            h0s = tf.tile(
                tf.reshape(self.h0, (1, self.num_units)), (n_batches, 1))
            h0s.set_shape((None, self.num_units))
            c0s = tf.tile(
                tf.reshape(self.c0, (1, self.num_units)), (n_batches, 1))
            c0s.set_shape((None, self.num_units))
            state = (c0s, h0s)
            if self.horizon is not None:
                outputs = []
                for idx in range(self.horizon):
                    output, state = self.lstm(
                        input[:, idx, :], state,
                        scope=self.scope)  # self.name)
                    outputs.append(tf.expand_dims(output, 1))
                outputs = tf.concat(axis=1, values=outputs)
                return outputs
            else:
                n_steps = input_shape[1]
                input = tf.reshape(input, tf.stack([n_batches, n_steps, -1]))
                # flatten extra dimensions
                shuffled_input = tf.transpose(input, (1, 0, 2))
                shuffled_input.set_shape((None, None, self.input_shape[-1]))
                hcs = tf.scan(
                    self.step,
                    elems=shuffled_input,
                    initializer=tf.concat(axis=1, values=[h0s, c0s]),
                )
                shuffled_hcs = tf.transpose(hcs, (1, 0, 2))
                shuffled_hs = shuffled_hcs[:, :, :self.num_units]
                return shuffled_hs

    def get_output_shape_for(self, input_shape):
        n_batch, n_steps = input_shape[:2]
        return n_batch, n_steps, self.num_units

    def get_step_layer(self, l_in, l_prev_state, name=None):
        return LSTMStepLayer(
            incomings=[l_in, l_prev_state], recurrent_layer=self, name=name)


def get_all_layers(layer, treat_as_input=None):
    """
    :type layer: Layer | list[Layer]
    :rtype: list[Layer]
    """
    # We perform a depth-first search. We add a layer to the result list only
    # after adding all its incoming layers (if any) or when detecting a cycle.
    # We use a LIFO stack to avoid ever running into recursion depth limits.
    try:
        queue = deque(layer)
    except TypeError:
        queue = deque([layer])
    seen = set()
    done = set()
    result = []

    # If treat_as_input is given, we pretend we've already collected all their
    # incoming layers.
    if treat_as_input is not None:
        seen.update(treat_as_input)

    while queue:
        # Peek at the leftmost node in the queue.
        layer = queue[0]
        if layer is None:
            # Some node had an input_layer set to `None`. Just ignore it.
            queue.popleft()
        elif layer not in seen:
            # We haven't seen this node yet: Mark it and queue all incomings
            # to be processed first. If there are no incomings, the node will
            # be appended to the result list in the next iteration.
            seen.add(layer)
            if hasattr(layer, 'input_layers'):
                queue.extendleft(reversed(layer.input_layers))
            elif hasattr(layer, 'input_layer'):
                queue.appendleft(layer.input_layer)
        else:
            # We've been here before: Either we've finished all its incomings,
            # or we've detected a cycle. In both cases, we remove the layer
            # from the queue and append it to the result list.
            queue.popleft()
            if layer not in done:
                result.append(layer)
                done.add(layer)

    return result


class NonlinearityLayer(Layer):
    def __init__(self, incoming, nonlinearity=tf.nn.relu, **kwargs):
        super(NonlinearityLayer, self).__init__(incoming, **kwargs)
        self.nonlinearity = (tf.identity
                             if nonlinearity is None else nonlinearity)

    def get_output_for(self, input, **kwargs):
        with tf.name_scope(self.name, values=[input]):
            return self.nonlinearity(input)

    def get_output_shape_for(self, input_shape):
        return input_shape


class BatchNormLayer(Layer):
    def __init__(self,
                 incoming,
                 center=True,
                 scale=False,
                 epsilon=0.001,
                 decay=0.9,
                 beta=tf.zeros_initializer(),
                 gamma=tf.ones_initializer(),
                 moving_mean=tf.zeros_initializer(),
                 moving_variance=tf.ones_initializer(),
                 **kwargs):
        super(BatchNormLayer, self).__init__(incoming, **kwargs)

        self.center = center
        self.scale = scale
        self.epsilon = epsilon
        self.decay = decay

        input_shape = incoming.output_shape
        axis = list(range(len(input_shape) - 1))
        params_shape = input_shape[-1:]

        if center:
            self.beta = self.add_param(
                beta,
                shape=params_shape,
                name='beta',
                trainable=True,
                regularizable=False)
        else:
            self.beta = None
        if scale:
            self.gamma = self.add_param(
                gamma,
                shape=params_shape,
                name='gamma',
                trainable=True,
                regularizable=True)
        else:
            self.gamma = None

        self.moving_mean = self.add_param(
            moving_mean,
            shape=params_shape,
            name='moving_mean',
            trainable=False,
            regularizable=False)
        self.moving_variance = self.add_param(
            moving_variance,
            shape=params_shape,
            name='moving_variance',
            trainable=False,
            regularizable=False)
        self.axis = axis

    def get_output_for(self, input, phase='train', **kwargs):
        with tf.name_scope(self.name, values=[input]):
            if phase == 'train':
                # Calculate the moments based on the individual batch.
                mean, variance = tf.nn.moments(
                    input, self.axis, shift=self.moving_mean)
                # Update the moving_mean and moving_variance moments.
                update_moving_mean = moving_averages.assign_moving_average(
                    self.moving_mean, mean, self.decay)
                update_moving_variance = moving_averages.assign_moving_average(
                    self.moving_variance, variance, self.decay)
                # Make sure the updates are computed here.
                with tf.control_dependencies(
                        [update_moving_mean,
                         update_moving_variance]):  # yapf:disable
                    output = tf.nn.batch_normalization(input, mean, variance,
                                                       self.beta, self.gamma,
                                                       self.epsilon)
            else:
                output = tf.nn.batch_normalization(
                    input, self.moving_mean, self.moving_variance, self.beta,
                    self.gamma, self.epsilon)
            output.set_shape(self.input_shape)
            return output

    def get_output_shape_for(self, input_shape):
        return input_shape


def batch_norm(layer, **kwargs):
    nonlinearity = getattr(layer, 'nonlinearity', None)
    scale = True
    if nonlinearity is not None:
        layer.nonlinearity = tf.identity
        if nonlinearity is tf.nn.relu:
            scale = False
    if hasattr(layer, 'b') and layer.b is not None:
        del layer.params[layer.b]
        layer.b = None
    bn_name = (kwargs.pop('name', None)
               or (getattr(layer, 'name', None) and layer.name + '_bn'))
    layer = BatchNormLayer(layer, name=bn_name, scale=scale, **kwargs)
    if nonlinearity is not None:
        nonlin_name = bn_name and bn_name + '_nonlin'
        layer = NonlinearityLayer(
            layer, nonlinearity=nonlinearity, name=nonlin_name)
    return layer


class ElemwiseSumLayer(MergeLayer):
    def __init__(self, incomings, **kwargs):
        super(ElemwiseSumLayer, self).__init__(incomings, **kwargs)

    def get_output_for(self, inputs, **kwargs):
        with tf.name_scope(self.name, values=[inputs]):
            return functools.reduce(tf.add, inputs)

    def get_output_shape_for(self, input_shapes):
        assert len(set(input_shapes)) == 1
        return input_shapes[0]


def get_output(layer_or_layers, inputs=None, **kwargs):
    # track accepted kwargs used by get_output_for
    accepted_kwargs = {'deterministic'}
    # obtain topological ordering of all layers the output layer(s) depend on
    treat_as_input = list(inputs.keys()) if isinstance(inputs, dict) else []
    all_layers = get_all_layers(layer_or_layers, treat_as_input)
    # initialize layer-to-expression mapping from all input layers
    all_outputs = dict(
        (layer, layer.input_var) for layer in all_layers
        if isinstance(layer, InputLayer) and layer not in treat_as_input)
    # update layer-to-expression mapping from given input(s), if any
    if isinstance(inputs, dict):
        all_outputs.update((layer, tf.convert_to_tensor(expr))
                           for layer, expr in list(inputs.items()))
    elif inputs is not None:
        if len(all_outputs) > 1:
            raise ValueError('get_output() was called with a single input '
                             'expression on a network with multiple input '
                             'layers. Please call it with a dictionary of '
                             'input expressions instead.')
        for input_layer in all_outputs:
            all_outputs[input_layer] = tf.convert_to_tensor(inputs)
    # update layer-to-expression mapping by propagating the inputs
    for layer in all_layers:
        if layer not in all_outputs:
            try:
                if isinstance(layer, MergeLayer):
                    layer_inputs = [
                        all_outputs[input_layer]
                        for input_layer in layer.input_layers
                    ]
                else:
                    layer_inputs = all_outputs[layer.input_layer]
            except KeyError:
                # one of the input_layer attributes must have been `None`
                raise ValueError(
                    'get_output() was called without giving an '
                    'input expression for the free-floating '
                    'layer %r. Please call it with a dictionary '
                    'mapping this layer to an input expression.' % layer)
            all_outputs[layer] = layer.get_output_for(layer_inputs, **kwargs)
            try:
                names, _, _, defaults, _, _, _ = inspect.getfullargspec(
                    layer.get_output_for)
            except TypeError:
                # If introspection is not possible, skip it
                pass
            else:
                if defaults is not None:
                    accepted_kwargs |= set(names[-len(defaults):])
            accepted_kwargs |= set(layer.get_output_kwargs)
    unused_kwargs = set(kwargs.keys()) - accepted_kwargs
    if unused_kwargs:
        suggestions = []
        for kwarg in unused_kwargs:
            suggestion = get_close_matches(kwarg, accepted_kwargs)
            if suggestion:
                suggestions.append(
                    '%s (perhaps you meant %s)' % (kwarg, suggestion[0]))
            else:
                suggestions.append(kwarg)
        warn('get_output() was called with unused kwargs:\n\t%s' %
             '\n\t'.join(suggestions))
    # return the output(s) of the requested layer(s) only
    try:
        return [all_outputs[layer] for layer in layer_or_layers]
    except TypeError:
        return all_outputs[layer_or_layers]


def unique(l):
    """Filters duplicates of iterable.
    Create a new list from l with duplicate entries removed,
    while preserving the original order.
    Parameters
    ----------
    l : iterable
        Input iterable to filter of duplicates.
    Returns
    -------
    list
        A list of elements of `l` without duplicates and in the same order.
    """
    new_list = []
    seen = set()
    for el in l:
        if el not in seen:
            new_list.append(el)
            seen.add(el)

    return new_list


def get_all_params(layer, **tags):
    """
    :type layer: Layer|list[Layer]
    """
    layers = get_all_layers(layer)
    params = chain.from_iterable(l.get_params(**tags) for l in layers)
    return unique(params)



#!/usr/bin/env python3
"""
This is an example to train a task with TRPO algorithm.

It uses an LSTM-based recurrent policy. To use a GRU-based recurrent
policy, swap the commented lines. Here it runs CartPole-v1 environment
with 100 iterations.

Results:
    AverageReturn: 100
    RiseTime: itr 13

"""
from garage.experiment import run_experiment
from garage.np.baselines import LinearFeatureBaseline
from garage.tf.algos import TRPO
import garage.tf.core.layers as L
from garage.tf.envs import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.optimizers import ConjugateGradientOptimizer
from garage.tf.optimizers import FiniteDifferenceHvp
from garage.tf.policies import CategoricalLSTMPolicy


def run_task(snapshot_config, *_):
    """Run task."""
    with LocalTFRunner(snapshot_config=snapshot_config) as runner:
        env = TfEnv(env_name='CartPole-v1')

        policy = CategoricalLSTMPolicy(
            name='policy',
            env_spec=env.spec,
            lstm_layer_cls=L.TfBasicLSTMLayer,
            # gru_layer_cls=L.GRULayer,
        )

        baseline = LinearFeatureBaseline(env_spec=env.spec)

        algo = TRPO(env_spec=env.spec,
                    policy=policy,
                    baseline=baseline,
                    max_path_length=100,
                    discount=0.99,
                    max_kl_step=0.01,
                    optimizer=ConjugateGradientOptimizer,
                    optimizer_args=dict(hvp_approach=FiniteDifferenceHvp(
                        base_eps=1e-5)))

        runner.setup(algo, env)
        runner.train(n_epochs=100, batch_size=4000)


run_experiment(
    run_task,
    snapshot_mode='last',
    seed=1,
)



"""
Unit test for ContinuousMLPQFunctionWithModel.

This test consists of four different ContinuousMLPQFunction: P1, P2, P3
and P4. P1 and P2 are from ContinuousMLPQFunction, which does not use
garage.tf.models.MLPMergeModel while P3 and P4 do use.

This test ensures the outputs from all the Q Functions are the same,
for the transition from using ContinuousMLPQFunction to
ContinuousMLPQFunctionWithModel.

It covers get_q_val and get_qval_sym
"""

from unittest import mock

import numpy as np
import tensorflow as tf

from garage.tf.envs import TfEnv
from garage.tf.q_functions import ContinuousMLPQFunction
from garage.tf.q_functions import ContinuousMLPQFunctionWithModel
from tests.fixtures import TfGraphTestCase
from tests.fixtures.envs.dummy import DummyBoxEnv


class TestContinuousMLPQFunctionTransit(TfGraphTestCase):
    def setup_method(self):
        with mock.patch('tensorflow.random.normal') as mock_rand:
            mock_rand.return_value = 0.5
            super().setup_method()
            self.obs_dim = (5, )
            self.act_dim = (2, )
            self.box_env = TfEnv(
                DummyBoxEnv(obs_dim=self.obs_dim, action_dim=self.act_dim))
            self.qf1 = ContinuousMLPQFunction(
                env_spec=self.box_env, hidden_sizes=(32, 32), name='QF1')
            self.qf2 = ContinuousMLPQFunction(
                env_spec=self.box_env, hidden_sizes=(64, 64), name='QF2')
            self.qf3 = ContinuousMLPQFunctionWithModel(
                env_spec=self.box_env, hidden_sizes=(32, 32), name='QF3')
            self.qf4 = ContinuousMLPQFunctionWithModel(
                env_spec=self.box_env, hidden_sizes=(64, 64), name='QF4')

            self.sess.run(tf.compat.v1.global_variables_initializer())

            for a, b in zip(self.qf3.get_trainable_vars(),
                            self.qf1.get_trainable_vars()):
                self.sess.run(a.assign(b))
            for a, b in zip(self.qf4.get_trainable_vars(),
                            self.qf2.get_trainable_vars()):
                self.sess.run(a.assign(b))

            self.obs = self.box_env.reset()
            self.act = np.full((2, ), 0.5)

    def test_get_qval(self):
        q_val1 = self.qf1.get_qval([self.obs], [self.act])
        q_val2 = self.qf2.get_qval([self.obs], [self.act])
        q_val3 = self.qf3.get_qval([self.obs], [self.act])
        q_val4 = self.qf4.get_qval([self.obs], [self.act])

        assert np.array_equal(q_val1, q_val3)
        assert np.array_equal(q_val2, q_val4)

        q_val1 = self.qf1.get_qval([self.obs, self.obs], [self.act, self.act])
        q_val2 = self.qf2.get_qval([self.obs, self.obs], [self.act, self.act])
        q_val3 = self.qf3.get_qval([self.obs, self.obs], [self.act, self.act])
        q_val4 = self.qf4.get_qval([self.obs, self.obs], [self.act, self.act])

        assert np.array_equal(q_val1, q_val3)
        assert np.allclose(q_val2, q_val4)

    def test_get_qval_sym(self):
        obs_ph = tf.compat.v1.placeholder(
            tf.float32, shape=(None, ) + self.obs_dim)
        act_ph = tf.compat.v1.placeholder(
            tf.float32, shape=(None, ) + self.act_dim)

        qval_sym1 = self.qf1.get_qval_sym(obs_ph, act_ph, name='qval_sym')
        qval_sym2 = self.qf2.get_qval_sym(obs_ph, act_ph, name='qval_sym')
        qval_sym3 = self.qf3.get_qval_sym(obs_ph, act_ph, name='qval_sym')
        qval_sym4 = self.qf4.get_qval_sym(obs_ph, act_ph, name='qval_sym')

        q_val1 = self.sess.run(
            qval_sym1, feed_dict={
                obs_ph: [self.obs],
                act_ph: [self.act]
            })
        q_val2 = self.sess.run(
            qval_sym2, feed_dict={
                obs_ph: [self.obs],
                act_ph: [self.act]
            })
        q_val3 = self.sess.run(
            qval_sym3, feed_dict={
                obs_ph: [self.obs],
                act_ph: [self.act]
            })
        q_val4 = self.sess.run(
            qval_sym4, feed_dict={
                obs_ph: [self.obs],
                act_ph: [self.act]
            })

        assert np.array_equal(q_val1, q_val3)
        assert np.array_equal(q_val2, q_val4)



import akro
import tensorflow as tf

from garage.core import Serializable
from garage.misc.overrides import overrides
from garage.tf.core import LayersPowered
import garage.tf.core.layers as L
from garage.tf.core.network import MLP
from garage.tf.misc import tensor_utils
from garage.tf.policies.base import Policy


class DeterministicMLPPolicy(Policy, LayersPowered, Serializable):
    def __init__(self,
                 env_spec,
                 name='DeterministicMLPPolicy',
                 hidden_sizes=(32, 32),
                 hidden_nonlinearity=tf.nn.relu,
                 output_nonlinearity=tf.nn.tanh,
                 prob_network=None,
                 bn=False):
        assert isinstance(env_spec.action_space, akro.Box)

        Serializable.quick_init(self, locals())

        self._prob_network_name = 'prob_network'
        with tf.compat.v1.variable_scope(name, 'DeterministicMLPPolicy'):
            if prob_network is None:
                prob_network = MLP(
                    input_shape=(env_spec.observation_space.flat_dim, ),
                    output_dim=env_spec.action_space.flat_dim,
                    hidden_sizes=hidden_sizes,
                    hidden_nonlinearity=hidden_nonlinearity,
                    output_nonlinearity=output_nonlinearity,
                    # batch_normalization=True,
                    name='mlp_prob_network',
                )

            with tf.name_scope(self._prob_network_name):
                prob_network_output = L.get_output(
                    prob_network.output_layer, deterministic=True)
            self._l_prob = prob_network.output_layer
            self._l_obs = prob_network.input_layer
            self._f_prob = tensor_utils.compile_function(
                [prob_network.input_layer.input_var], prob_network_output)

        self.prob_network = prob_network
        self.name = name

        # Note the deterministic=True argument. It makes sure that when getting
        # actions from single observations, we do not update params in the
        # batch normalization layers.
        # TODO: this doesn't currently work properly in the tf version so we
        # leave out batch_norm
        super(DeterministicMLPPolicy, self).__init__(env_spec)
        LayersPowered.__init__(self, [prob_network.output_layer])

    @property
    def vectorized(self):
        return True

    @overrides
    def get_action(self, observation):
        flat_obs = self.observation_space.flatten(observation)
        action = self._f_prob([flat_obs])[0]
        return action, dict()

    @overrides
    def get_actions(self, observations):
        flat_obs = self.observation_space.flatten_n(observations)
        actions = self._f_prob(flat_obs)
        return actions, dict()

    def get_action_sym(self, obs_var, name=None):
        with tf.name_scope(name, 'get_action_sym', values=[obs_var]):
            with tf.name_scope(self._prob_network_name, values=[obs_var]):
                return L.get_output(self.prob_network.output_layer, obs_var)



import akro
import numpy as np
import tensorflow as tf

from garage.core import Serializable
from garage.misc.overrides import overrides
from garage.tf.core import LayersPowered
import garage.tf.core.layers as L
from garage.tf.core.network import LSTMNetwork
from garage.tf.distributions import RecurrentDiagonalGaussian
from garage.tf.misc import tensor_utils
from garage.tf.policies.base import StochasticPolicy


class GaussianLSTMPolicy(StochasticPolicy, LayersPowered, Serializable):
    def __init__(
            self,
            env_spec,
            name='GaussianLSTMPolicy',
            hidden_dim=32,
            hidden_nonlinearity=tf.tanh,
            recurrent_nonlinearity=tf.nn.sigmoid,
            recurrent_w_x_init=L.XavierUniformInitializer(),
            recurrent_w_h_init=L.OrthogonalInitializer(),
            output_nonlinearity=None,
            output_w_init=L.XavierUniformInitializer(),
            feature_network=None,
            state_include_action=True,
            learn_std=True,
            init_std=1.0,
            lstm_layer_cls=L.LSTMLayer,
            use_peepholes=False,
            std_share_network=False,
    ):
        """
        :param env_spec: A spec for the env.
        :param hidden_dim: dimension of hidden layer
        :param hidden_nonlinearity: nonlinearity used for each hidden layer
        :return:
        """
        assert isinstance(env_spec.action_space, akro.Box)

        self._mean_network_name = 'mean_network'
        self._std_network_name = 'std_network'
        with tf.compat.v1.variable_scope(name, 'GaussianLSTMPolicy'):
            Serializable.quick_init(self, locals())
            super(GaussianLSTMPolicy, self).__init__(env_spec)

            obs_dim = env_spec.observation_space.flat_dim
            action_dim = env_spec.action_space.flat_dim

            if state_include_action:
                input_dim = obs_dim + action_dim
            else:
                input_dim = obs_dim

            l_input = L.InputLayer(shape=(None, None, input_dim), name='input')

            if feature_network is None:
                feature_dim = input_dim
                l_flat_feature = None
                l_feature = l_input
            else:
                feature_dim = feature_network.output_layer.output_shape[-1]
                l_flat_feature = feature_network.output_layer
                l_feature = L.OpLayer(
                    l_flat_feature,
                    extras=[l_input],
                    name='reshape_feature',
                    op=lambda flat_feature, input: tf.reshape(
                        flat_feature,
                        tf.stack([
                            tf.shape(input)[0],
                            tf.shape(input)[1], feature_dim
                        ])),
                    shape_op=lambda _, input_shape: (input_shape[
                        0], input_shape[1], feature_dim))

            if std_share_network:
                mean_network = LSTMNetwork(
                    input_shape=(feature_dim, ),
                    input_layer=l_feature,
                    output_dim=2 * action_dim,
                    hidden_dim=hidden_dim,
                    hidden_nonlinearity=hidden_nonlinearity,
                    recurrent_nonlinearity=recurrent_nonlinearity,
                    recurrent_w_x_init=recurrent_w_x_init,
                    recurrent_w_h_init=recurrent_w_h_init,
                    output_nonlinearity=output_nonlinearity,
                    output_w_init=output_w_init,
                    lstm_layer_cls=lstm_layer_cls,
                    name='lstm_mean_network',
                    use_peepholes=use_peepholes,
                )

                l_mean = L.SliceLayer(
                    mean_network.output_layer,
                    slice(action_dim),
                    name='mean_slice',
                )

                l_step_mean = L.SliceLayer(
                    mean_network.step_output_layer,
                    slice(action_dim),
                    name='step_mean_slice',
                )

                l_log_std = L.SliceLayer(
                    mean_network.output_layer,
                    slice(action_dim, 2 * action_dim),
                    name='log_std_slice',
                )

                l_step_log_std = L.SliceLayer(
                    mean_network.step_output_layer,
                    slice(action_dim, 2 * action_dim),
                    name='step_log_std_slice',
                )
            else:
                mean_network = LSTMNetwork(
                    input_shape=(feature_dim, ),
                    input_layer=l_feature,
                    output_dim=action_dim,
                    hidden_dim=hidden_dim,
                    hidden_nonlinearity=hidden_nonlinearity,
                    recurrent_nonlinearity=recurrent_nonlinearity,
                    recurrent_w_x_init=recurrent_w_x_init,
                    recurrent_w_h_init=recurrent_w_h_init,
                    output_nonlinearity=output_nonlinearity,
                    output_w_init=output_w_init,
                    lstm_layer_cls=lstm_layer_cls,
                    name='lstm_mean_network',
                    use_peepholes=use_peepholes,
                )

                l_mean = mean_network.output_layer

                l_step_mean = mean_network.step_output_layer

                l_log_std = L.ParamLayer(
                    mean_network.input_layer,
                    num_units=action_dim,
                    param=tf.constant_initializer(np.log(init_std)),
                    name='output_log_std',
                    trainable=learn_std,
                )

                l_step_log_std = L.ParamLayer(
                    mean_network.step_input_layer,
                    num_units=action_dim,
                    param=l_log_std.param,
                    name='step_output_log_std',
                    trainable=learn_std,
                )

            self.mean_network = mean_network
            self.feature_network = feature_network
            self.l_input = l_input
            self.state_include_action = state_include_action
            self.name = name

            flat_input_var = tf.compat.v1.placeholder(
                dtype=tf.float32, shape=(None, input_dim), name='flat_input')
            if feature_network is None:
                feature_var = flat_input_var
            else:
                feature_var = L.get_output(
                    l_flat_feature,
                    {feature_network.input_layer: flat_input_var})

            with tf.name_scope(self._mean_network_name, values=[feature_var]):
                (out_step_mean, out_step_hidden, out_mean_cell) = L.get_output(
                    [
                        l_step_mean, mean_network.step_hidden_layer,
                        mean_network.step_cell_layer
                    ], {mean_network.step_input_layer: feature_var})
                out_step_mean = tf.identity(out_step_mean, 'step_mean')
                out_step_hidden = tf.identity(out_step_hidden, 'step_hidden')
                out_mean_cell = tf.identity(out_mean_cell, 'mean_cell')

            with tf.name_scope(self._std_network_name, values=[feature_var]):
                out_step_log_std = L.get_output(
                    l_step_log_std,
                    {mean_network.step_input_layer: feature_var})
                out_step_log_std = tf.identity(out_step_log_std,
                                               'step_log_std')

            self.f_step_mean_std = tensor_utils.compile_function([
                flat_input_var,
                mean_network.step_prev_state_layer.input_var,
            ], [
                out_step_mean, out_step_log_std, out_step_hidden, out_mean_cell
            ])

            self.l_mean = l_mean
            self.l_log_std = l_log_std

            self.input_dim = input_dim
            self.action_dim = action_dim
            self.hidden_dim = hidden_dim

            self.prev_actions = None
            self.prev_hiddens = None
            self.prev_cells = None
            self.dist = RecurrentDiagonalGaussian(action_dim)

            out_layers = [l_mean, l_log_std]
            if feature_network is not None:
                out_layers.append(feature_network.output_layer)

            LayersPowered.__init__(self, out_layers)

    @overrides
    def dist_info_sym(self, obs_var, state_info_vars, name=None):
        with tf.name_scope(name, 'dist_info_sym', [obs_var, state_info_vars]):
            n_batches = tf.shape(obs_var)[0]
            n_steps = tf.shape(obs_var)[1]
            obs_var = tf.reshape(obs_var, tf.stack([n_batches, n_steps, -1]))
            if self.state_include_action:
                prev_action_var = state_info_vars['prev_action']
                all_input_var = tf.concat(
                    axis=2, values=[obs_var, prev_action_var])
            else:
                all_input_var = obs_var
            if self.feature_network is None:
                with tf.name_scope(
                        self._mean_network_name, values=[all_input_var]):
                    means = L.get_output(self.mean_network.output_layer,
                                         {self.l_input: all_input_var})
                with tf.name_scope(
                        self._std_network_name, values=[all_input_var]):
                    log_stds = L.get_output(self.l_log_std,
                                            {self.l_input: all_input_var})
            else:
                flat_input_var = tf.reshape(all_input_var,
                                            (-1, self.input_dim))
                with tf.name_scope(
                        self._mean_network_name,
                        values=[all_input_var, flat_input_var]):
                    means = L.get_output(
                        self.mean_network.output_layer, {
                            self.l_input: all_input_var,
                            self.feature_network.input_layer: flat_input_var
                        })
                with tf.name_scope(
                        self._std_network_name,
                        values=[all_input_var, flat_input_var]):
                    log_stds = L.get_output(
                        self.l_log_std, {
                            self.l_input: all_input_var,
                            self.feature_network.input_layer: flat_input_var
                        })
            return dict(mean=means, log_std=log_stds)

    @property
    def vectorized(self):
        return True

    def reset(self, dones=None):
        if dones is None:
            dones = [True]
        dones = np.asarray(dones)
        if self.prev_actions is None or len(dones) != len(self.prev_actions):
            self.prev_actions = np.zeros((len(dones),
                                          self.action_space.flat_dim))
            self.prev_hiddens = np.zeros((len(dones), self.hidden_dim))
            self.prev_cells = np.zeros((len(dones), self.hidden_dim))

        self.prev_actions[dones] = 0.
        self.prev_hiddens[dones] = self.mean_network.hid_init_param.eval()
        self.prev_cells[dones] = self.mean_network.cell_init_param.eval()

    # The return value is a pair. The first item is a matrix (N, A), where each
    # entry corresponds to the action value taken. The second item is a vector
    # of length N, where each entry is the density value for that action, under
    # the current policy
    @overrides
    def get_action(self, observation):
        actions, agent_infos = self.get_actions([observation])
        return actions[0], {k: v[0] for k, v in agent_infos.items()}

    @overrides
    def get_actions(self, observations):
        flat_obs = self.observation_space.flatten_n(observations)
        if self.state_include_action:
            assert self.prev_actions is not None
            all_input = np.concatenate([flat_obs, self.prev_actions], axis=-1)
        else:
            all_input = flat_obs
        # probs, hidden_vec, cell_vec = self.f_step_prob(
        #     all_input, self.prev_hiddens, self.prev_cells)
        means, log_stds, hidden_vec, cell_vec = self.f_step_mean_std(
            all_input, np.hstack([self.prev_hiddens, self.prev_cells]))
        rnd = np.random.normal(size=means.shape)
        actions = rnd * np.exp(log_stds) + means
        prev_actions = self.prev_actions
        self.prev_actions = self.action_space.flatten_n(actions)
        self.prev_hiddens = hidden_vec
        self.prev_cells = cell_vec
        agent_info = dict(mean=means, log_std=log_stds)
        if self.state_include_action:
            agent_info['prev_action'] = np.copy(prev_actions)
        return actions, agent_info

    @property
    @overrides
    def recurrent(self):
        return True

    @property
    def distribution(self):
        return self.dist

    @property
    def state_info_specs(self):
        if self.state_include_action:
            return [
                ('prev_action', (self.action_dim, )),
            ]
        else:
            return []



"""Gaussian Conv Regressor."""
from dowel import tabular
import numpy as np
import tensorflow as tf

from garage.core import Serializable
from garage.tf.core import LayersPowered
from garage.tf.core import Parameterized
import garage.tf.core.layers as layers
from garage.tf.core.network import ConvNetwork
from garage.tf.distributions import DiagonalGaussian
from garage.tf.misc import tensor_utils
from garage.tf.optimizers import LbfgsOptimizer
from garage.tf.optimizers import PenaltyLbfgsOptimizer


class GaussianConvRegressor(LayersPowered, Serializable, Parameterized):
    """
    A regressor fits a Gaussian distribution to outputs using a ConvNetwork.

    Args:
        Most of the net param infos can be found at the docstring of
         garage.tf.core.network.ConvNetwork.
        input_shape: a tuple or list contains input shapes of the mean net.
        conv_filters: a list of numbers of the mean convnet kernels.
        conv_filter_sizes: a list of sizes of the mean convnet kernels.
        conv_strides: a list of strides of the mean convnet kernels.
        conv_pads: a list of pad formats of the mean convnet.
        hidden_sizes: a list of numbers of hidden units for all fc layers.
        hidden_nonlinearity: a nonlinearity from tf.nn, shared by all conv and
         fc layers.

    """

    def __init__(self,
                 input_shape,
                 output_dim,
                 conv_filters,
                 conv_filter_sizes,
                 conv_strides,
                 conv_pads,
                 hidden_sizes,
                 hidden_nonlinearity=tf.nn.tanh,
                 output_nonlinearity=None,
                 name='GaussianConvRegressor',
                 mean_network=None,
                 learn_std=True,
                 init_std=1.0,
                 adaptive_std=False,
                 std_share_network=False,
                 std_conv_filters=[],
                 std_conv_filter_sizes=[],
                 std_conv_strides=[],
                 std_conv_pads=[],
                 std_hidden_sizes=[],
                 std_hidden_nonlinearity=None,
                 std_output_nonlinearity=None,
                 normalize_inputs=True,
                 normalize_outputs=True,
                 subsample_factor=1.,
                 optimizer=None,
                 optimizer_args=dict(),
                 use_trust_region=True,
                 max_kl_step=0.01):
        Parameterized.__init__(self)
        Serializable.quick_init(self, locals())
        self._mean_network_name = 'mean_network'
        self._std_network_name = 'std_network'

        with tf.compat.v1.variable_scope(name):
            if optimizer is None:
                if use_trust_region:
                    optimizer = PenaltyLbfgsOptimizer(**optimizer_args)
                else:
                    optimizer = LbfgsOptimizer(**optimizer_args)
            else:
                optimizer = optimizer(**optimizer_args)

            self._optimizer = optimizer
            self._subsample_factor = subsample_factor

            if mean_network is None:
                if std_share_network:
                    b = np.concatenate(
                        [
                            np.zeros(output_dim),
                            np.full(output_dim, np.log(init_std))
                        ],
                        axis=0)  # yapf: disable
                    b = tf.constant_initializer(b)
                    mean_network = ConvNetwork(
                        name=self._mean_network_name,
                        input_shape=input_shape,
                        output_dim=2 * output_dim,
                        conv_filters=conv_filters,
                        conv_filter_sizes=conv_filter_sizes,
                        conv_strides=conv_strides,
                        conv_pads=conv_pads,
                        hidden_sizes=hidden_sizes,
                        hidden_nonlinearity=hidden_nonlinearity,
                        output_nonlinearity=output_nonlinearity,
                        output_b_init=b)
                    l_mean = layers.SliceLayer(
                        mean_network.output_layer,
                        slice(output_dim),
                        name='mean_slice',
                    )
                else:
                    mean_network = ConvNetwork(
                        name=self._mean_network_name,
                        input_shape=input_shape,
                        output_dim=output_dim,
                        conv_filters=conv_filters,
                        conv_filter_sizes=conv_filter_sizes,
                        conv_strides=conv_strides,
                        conv_pads=conv_pads,
                        hidden_sizes=hidden_sizes,
                        hidden_nonlinearity=hidden_nonlinearity,
                        output_nonlinearity=output_nonlinearity)
                    l_mean = mean_network.output_layer

            if adaptive_std:
                l_log_std = ConvNetwork(
                    name=self._std_network_name,
                    input_shape=input_shape,
                    output_dim=output_dim,
                    conv_filters=std_conv_filters,
                    conv_filter_sizes=std_conv_filter_sizes,
                    conv_strides=std_conv_strides,
                    conv_pads=std_conv_pads,
                    hidden_sizes=std_hidden_sizes,
                    hidden_nonlinearity=std_hidden_nonlinearity,
                    output_nonlinearity=std_output_nonlinearity,
                    output_b_init=tf.constant_initializer(np.log(init_std)),
                ).output_layer
            elif std_share_network:
                l_log_std = layers.SliceLayer(
                    mean_network.output_layer,
                    slice(output_dim, 2 * output_dim),
                    name='log_std_slice',
                )
            else:
                l_log_std = layers.ParamLayer(
                    mean_network.input_layer,
                    num_units=output_dim,
                    param=tf.constant_initializer(np.log(init_std)),
                    trainable=learn_std,
                    name=self._std_network_name,
                )

            LayersPowered.__init__(self, [l_mean, l_log_std])

            xs_var = mean_network.input_layer.input_var
            ys_var = tf.compat.v1.placeholder(
                dtype=tf.float32, name='ys', shape=(None, output_dim))
            old_means_var = tf.compat.v1.placeholder(
                dtype=tf.float32, name='ys', shape=(None, output_dim))
            old_log_stds_var = tf.compat.v1.placeholder(
                dtype=tf.float32,
                name='old_log_stds',
                shape=(None, output_dim))

            x_mean_var = tf.Variable(
                np.zeros((1, np.prod(input_shape)), dtype=np.float32),
                name='x_mean',
            )
            x_std_var = tf.Variable(
                np.ones((1, np.prod(input_shape)), dtype=np.float32),
                name='x_std',
            )
            y_mean_var = tf.Variable(
                np.zeros((1, output_dim), dtype=np.float32),
                name='y_mean',
            )
            y_std_var = tf.Variable(
                np.ones((1, output_dim), dtype=np.float32),
                name='y_std',
            )

            normalized_xs_var = (xs_var - x_mean_var) / x_std_var
            normalized_ys_var = (ys_var - y_mean_var) / y_std_var

            with tf.name_scope(
                    self._mean_network_name, values=[normalized_xs_var]):
                normalized_means_var = layers.get_output(
                    l_mean, {mean_network.input_layer: normalized_xs_var})
            with tf.name_scope(
                    self._std_network_name, values=[normalized_xs_var]):
                normalized_log_stds_var = layers.get_output(
                    l_log_std, {mean_network.input_layer: normalized_xs_var})

            means_var = normalized_means_var * y_std_var + y_mean_var
            log_stds_var = normalized_log_stds_var + tf.math.log(y_std_var)

            normalized_old_means_var = (old_means_var - y_mean_var) / y_std_var
            normalized_old_log_stds_var = (
                old_log_stds_var - tf.math.log(y_std_var))

            dist = self._dist = DiagonalGaussian(output_dim)

            normalized_dist_info_vars = dict(
                mean=normalized_means_var, log_std=normalized_log_stds_var)

            mean_kl = tf.reduce_mean(
                dist.kl_sym(
                    dict(
                        mean=normalized_old_means_var,
                        log_std=normalized_old_log_stds_var),
                    normalized_dist_info_vars,
                ))

            loss = -tf.reduce_mean(
                dist.log_likelihood_sym(normalized_ys_var,
                                        normalized_dist_info_vars))

            self._f_predict = tensor_utils.compile_function([xs_var],
                                                            means_var)
            self._f_pdists = tensor_utils.compile_function(
                [xs_var], [means_var, log_stds_var])
            self._l_mean = l_mean
            self._l_log_std = l_log_std

            optimizer_args = dict(
                loss=loss,
                target=self,
                network_outputs=[
                    normalized_means_var, normalized_log_stds_var
                ],
            )

            if use_trust_region:
                optimizer_args['leq_constraint'] = (mean_kl, max_kl_step)
                optimizer_args['inputs'] = [
                    xs_var, ys_var, old_means_var, old_log_stds_var
                ]
            else:
                optimizer_args['inputs'] = [xs_var, ys_var]

            self._optimizer.update_opt(**optimizer_args)

            self._use_trust_region = use_trust_region
            self._name = name

            self._normalize_inputs = normalize_inputs
            self._normalize_outputs = normalize_outputs
            self._mean_network = mean_network
            self._x_mean_var = x_mean_var
            self._x_std_var = x_std_var
            self._y_mean_var = y_mean_var
            self._y_std_var = y_std_var

    def fit(self, xs, ys):
        """Optimize the regressor based on the inputs."""
        if self._subsample_factor < 1:
            num_samples_tot = xs.shape[0]
            idx = np.random.randint(
                0, num_samples_tot,
                int(num_samples_tot * self._subsample_factor))
            xs, ys = xs[idx], ys[idx]

        sess = tf.compat.v1.get_default_session()
        if self._normalize_inputs:
            # recompute normalizing constants for inputs
            sess.run([
                tf.compat.v1.assign(self._x_mean_var,
                                    np.mean(xs, axis=0, keepdims=True)),
                tf.compat.v1.assign(self._x_std_var,
                                    np.std(xs, axis=0, keepdims=True) + 1e-8),
            ])
        if self._normalize_outputs:
            # recompute normalizing constants for outputs
            sess.run([
                tf.compat.v1.assign(self._y_mean_var,
                                    np.mean(ys, axis=0, keepdims=True)),
                tf.compat.v1.assign(self._y_std_var,
                                    np.std(ys, axis=0, keepdims=True) + 1e-8),
            ])
        if self._use_trust_region:
            old_means, old_log_stds = self._f_pdists(xs)
            inputs = [xs, ys, old_means, old_log_stds]
        else:
            inputs = [xs, ys]
        loss_before = self._optimizer.loss(inputs)
        if self._name:
            prefix = self._name + '/'
        else:
            prefix = ''
        tabular.record(prefix + 'LossBefore', loss_before)
        self._optimizer.optimize(inputs)
        loss_after = self._optimizer.loss(inputs)
        tabular.record(prefix + 'LossAfter', loss_after)
        if self._use_trust_region:
            tabular.record(prefix + 'MeanKL',
                           self._optimizer.constraint_val(inputs))
        tabular.record(prefix + 'dLoss', loss_before - loss_after)

    def predict(self, xs):
        """
        Return the maximum likelihood estimate of the predicted y.

        :param xs:
        :return:

        """
        return self._f_predict(xs)

    def sample_predict(self, xs):
        """
        Sample one possible output from the prediction distribution.

        :param xs:
        :return:

        """
        means, log_stds = self._f_pdists(xs)
        return self._dist.sample(dict(mean=means, log_std=log_stds))

    def predict_log_likelihood(self, xs, ys):
        """
        Predict the log likelihood of the distribution.

        :param xs:
        :param ys:
        :return:

        """
        means, log_stds = self._f_pdists(xs)
        return self._dist.log_likelihood(ys, dict(
            mean=means, log_std=log_stds))

    def log_likelihood_sym(self, x_var, y_var, name=None):
        """
        Return the symbolic log likelihood info of the distribution.

        :param x_var:
        :param y_var:
        :param name:

        """
        with tf.name_scope(name, 'log_likelihood_sym', [x_var, y_var]):
            normalized_xs_var = (x_var - self._x_mean_var) / self._x_std_var

            with tf.name_scope(
                    self._mean_network_name, values=[normalized_xs_var]):
                normalized_means_var = layers.get_output(
                    self._l_mean,
                    {self._mean_network.input_layer: normalized_xs_var})
            with tf.name_scope(
                    self._std_network_name, values=[normalized_xs_var]):
                normalized_log_stds_var = layers.get_output(
                    self._l_log_std,
                    {self._mean_network.input_layer: normalized_xs_var})

            means_var = (
                normalized_means_var * self._y_std_var + self._y_mean_var)
            log_stds_var = normalized_log_stds_var + tf.math.log(
                self._y_std_var)

            return self._dist.log_likelihood_sym(
                y_var, dict(mean=means_var, log_std=log_stds_var))



"""
Unit test for ContinuousMLPPolicyWithModel.

This test consists of two ContinuousMLPPolicy: P1 and P2, and
two ContinuousMLPPolicyWithModel: P3 and P4.

This test ensures the outputs from all the policies are the same,
for the transition from using ContinuousMLPPolicy to
ContinuousMLPPolicyWithModel.

Notice that ContinuousMLPPolicy scales output according to
action space boundary while ContinuousMLPPolicyWithModel doesn't.

It covers get_action, get_actions, get_action_sym.
"""
from unittest import mock

import numpy as np
import tensorflow as tf

from garage.tf.envs import TfEnv
from garage.tf.policies import ContinuousMLPPolicy
from garage.tf.policies import ContinuousMLPPolicyWithModel
from tests.fixtures import TfGraphTestCase
from tests.fixtures.envs.dummy import DummyBoxEnv


class TestContinuousMLPPolicyWithModelTransit(TfGraphTestCase):
    def setup_method(self):
        with mock.patch('tensorflow.random.normal') as mock_rand:
            mock_rand.return_value = 0.5
            super().setup_method()
            self.box_env = TfEnv(DummyBoxEnv())
            self.policy1 = ContinuousMLPPolicy(
                env_spec=self.box_env, hidden_sizes=(32, 32), name='P1')
            self.policy2 = ContinuousMLPPolicy(
                env_spec=self.box_env, hidden_sizes=(64, 64), name='P2')
            self.policy3 = ContinuousMLPPolicyWithModel(
                env_spec=self.box_env, hidden_sizes=(32, 32), name='P3')
            self.policy4 = ContinuousMLPPolicyWithModel(
                env_spec=self.box_env, hidden_sizes=(64, 64), name='P4')

            self.sess.run(tf.compat.v1.global_variables_initializer())
            for a, b in zip(self.policy3.get_params(),
                            self.policy1.get_params()):
                self.sess.run(a.assign(b))
            for a, b in zip(self.policy4.get_params(),
                            self.policy2.get_params()):
                self.sess.run(a.assign(b))

            self.obs = self.box_env.reset()
            self.action_bound = self.box_env.action_space.high
            assert self.policy1.vectorized == self.policy2.vectorized
            assert self.policy3.vectorized == self.policy4.vectorized

    @mock.patch('numpy.random.normal')
    def test_get_action(self, mock_rand):
        mock_rand.return_value = 0.5
        action1, _ = self.policy1.get_action(self.obs)
        action2, _ = self.policy2.get_action(self.obs)
        action3, _ = self.policy3.get_action(self.obs)
        action4, _ = self.policy4.get_action(self.obs)

        assert np.array_equal(action1, action3 * self.action_bound)
        assert np.array_equal(action2, action4 * self.action_bound)

        actions1, _ = self.policy1.get_actions([self.obs, self.obs])
        actions2, _ = self.policy2.get_actions([self.obs, self.obs])
        actions3, _ = self.policy3.get_actions([self.obs, self.obs])
        actions4, _ = self.policy4.get_actions([self.obs, self.obs])

        assert np.array_equal(actions1, actions3 * self.action_bound)
        assert np.array_equal(actions2, actions4 * self.action_bound)

    def test_get_action_sym(self):
        obs_dim = self.box_env.spec.observation_space.flat_dim
        state_input = tf.compat.v1.placeholder(
            tf.float32, shape=(None, obs_dim))

        action_sym1 = self.policy1.get_action_sym(
            state_input, name='action_sym')
        action_sym2 = self.policy2.get_action_sym(
            state_input, name='action_sym')
        action_sym3 = self.policy3.get_action_sym(
            state_input, name='action_sym')
        action_sym4 = self.policy4.get_action_sym(
            state_input, name='action_sym')

        action1 = self.sess.run(
            action_sym1, feed_dict={state_input: [self.obs]})
        action2 = self.sess.run(
            action_sym2, feed_dict={state_input: [self.obs]})
        action3 = self.sess.run(
            action_sym3, feed_dict={state_input: [self.obs]})
        action4 = self.sess.run(
            action_sym4, feed_dict={state_input: [self.obs]})

        assert np.array_equal(action1, action3 * self.action_bound)
        assert np.array_equal(action2, action4 * self.action_bound)



import numpy as np
import tensorflow as tf

from garage.core import Serializable
from garage.tf.core import layers as ly
from garage.tf.core import LayersPowered


class MLP(LayersPowered, Serializable):
    def __init__(
            self,
            output_dim,
            hidden_sizes,
            hidden_nonlinearity,
            output_nonlinearity,
            name=None,
            hidden_w_init=ly.XavierUniformInitializer(),
            hidden_b_init=tf.zeros_initializer(),
            output_w_init=ly.XavierUniformInitializer(),
            output_b_init=tf.zeros_initializer(),
            input_var=None,
            input_layer=None,
            input_shape=None,
            batch_normalization=False,
            weight_normalization=False,
    ):

        Serializable.quick_init(self, locals())

        with tf.compat.v1.variable_scope(name, 'MLP'):
            if input_layer is None:
                l_in = ly.InputLayer(
                    shape=(None, ) + input_shape,
                    input_var=input_var,
                    name='input')
            else:
                l_in = input_layer
            self._layers = [l_in]
            l_hid = l_in
            if batch_normalization:
                l_hid = ly.batch_norm(l_hid)
            for idx, hidden_size in enumerate(hidden_sizes):
                l_hid = ly.DenseLayer(
                    l_hid,
                    num_units=hidden_size,
                    nonlinearity=hidden_nonlinearity,
                    name='hidden_%d' % idx,
                    w=hidden_w_init,
                    b=hidden_b_init,
                    weight_normalization=weight_normalization)
                if batch_normalization:
                    l_hid = ly.batch_norm(l_hid)
                self._layers.append(l_hid)
            l_out = ly.DenseLayer(
                l_hid,
                num_units=output_dim,
                nonlinearity=output_nonlinearity,
                name='output',
                w=output_w_init,
                b=output_b_init,
                weight_normalization=weight_normalization)
            if batch_normalization:
                l_out = ly.batch_norm(l_out)
            self._layers.append(l_out)
            self._l_in = l_in
            self._l_out = l_out

            LayersPowered.__init__(self, l_out)

    @property
    def input_layer(self):
        return self._l_in

    @property
    def output_layer(self):
        return self._l_out

    @property
    def input_var(self):
        return self._l_in.input_var

    @property
    def layers(self):
        return self._layers


class ConvNetwork(LayersPowered, Serializable):
    def __init__(self,
                 input_shape,
                 output_dim,
                 conv_filters,
                 conv_filter_sizes,
                 conv_strides,
                 conv_pads,
                 hidden_sizes,
                 hidden_nonlinearity,
                 output_nonlinearity,
                 name=None,
                 hidden_w_init=ly.XavierUniformInitializer(),
                 hidden_b_init=tf.zeros_initializer(),
                 output_w_init=ly.XavierUniformInitializer(),
                 output_b_init=tf.zeros_initializer(),
                 input_var=None,
                 input_layer=None,
                 batch_normalization=False,
                 weight_normalization=False):
        Serializable.quick_init(self, locals())
        """
        A network composed of several convolution layers followed by some fc
        layers.
        input_shape: (width,height,channel)
            HOWEVER, network inputs are assumed flattened. This network will
            first unflatten the inputs and then apply the standard convolutions
            and so on.
        conv_filters: a list of numbers of convolution kernel
        conv_filter_sizes: a list of sizes (int) of the convolution kernels
        conv_strides: a list of strides (int) of the conv kernels
        conv_pads: a list of pad formats (either 'SAME' or 'VALID')
        hidden_nonlinearity: a nonlinearity from tf.nn, shared by all conv and
         fc layers
        hidden_sizes: a list of numbers of hidden units for all fc layers
        """
        with tf.compat.v1.variable_scope(name, 'ConvNetwork'):
            if input_layer is not None:
                l_in = input_layer
                l_hid = l_in
            elif len(input_shape) == 3:
                l_in = ly.InputLayer(
                    shape=(None, np.prod(input_shape)),
                    input_var=input_var,
                    name='input')
                l_hid = ly.reshape(
                    l_in, ([0], ) + input_shape, name='reshape_input')
            elif len(input_shape) == 2:
                l_in = ly.InputLayer(
                    shape=(None, np.prod(input_shape)),
                    input_var=input_var,
                    name='input')
                input_shape = (1, ) + input_shape
                l_hid = ly.reshape(
                    l_in, ([0], ) + input_shape, name='reshape_input')
            else:
                l_in = ly.InputLayer(
                    shape=(None, ) + input_shape,
                    input_var=input_var,
                    name='input')
                l_hid = l_in

            if batch_normalization:
                l_hid = ly.batch_norm(l_hid)
            for idx, conv_filter, filter_size, stride, pad in zip(
                    range(len(conv_filters)),
                    conv_filters,
                    conv_filter_sizes,
                    conv_strides,
                    conv_pads,
            ):
                l_hid = ly.Conv2DLayer(
                    l_hid,
                    num_filters=conv_filter,
                    filter_size=filter_size,
                    stride=(stride, stride),
                    pad=pad,
                    nonlinearity=hidden_nonlinearity,
                    name='conv_hidden_%d' % idx,
                    weight_normalization=weight_normalization,
                )
                if batch_normalization:
                    l_hid = ly.batch_norm(l_hid)

            if output_nonlinearity == ly.spatial_expected_softmax:
                assert not hidden_sizes
                assert output_dim == conv_filters[-1] * 2
                l_hid.nonlinearity = tf.identity
                l_out = ly.SpatialExpectedSoftmaxLayer(l_hid)
            else:
                l_hid = ly.flatten(l_hid, name='conv_flatten')
                for idx, hidden_size in enumerate(hidden_sizes):
                    l_hid = ly.DenseLayer(
                        l_hid,
                        num_units=hidden_size,
                        nonlinearity=hidden_nonlinearity,
                        name='hidden_%d' % idx,
                        w=hidden_w_init,
                        b=hidden_b_init,
                        weight_normalization=weight_normalization,
                    )
                    if batch_normalization:
                        l_hid = ly.batch_norm(l_hid)
                l_out = ly.DenseLayer(
                    l_hid,
                    num_units=output_dim,
                    nonlinearity=output_nonlinearity,
                    name='output',
                    w=output_w_init,
                    b=output_b_init,
                    weight_normalization=weight_normalization,
                )
                if batch_normalization:
                    l_out = ly.batch_norm(l_out)
            self._l_in = l_in
            self._l_out = l_out
            # self._input_var = l_in.input_var

        LayersPowered.__init__(self, l_out)

    @property
    def input_layer(self):
        return self._l_in

    @property
    def output_layer(self):
        return self._l_out

    @property
    def input_var(self):
        return self._l_in.input_var


class GRUNetwork:
    def __init__(self,
                 input_shape,
                 output_dim,
                 hidden_dim,
                 name=None,
                 hidden_nonlinearity=tf.nn.relu,
                 output_w_init=ly.XavierUniformInitializer(),
                 recurrent_nonlinearity=tf.nn.sigmoid,
                 recurrent_w_x_init=ly.XavierUniformInitializer(),
                 recurrent_w_h_init=ly.OrthogonalInitializer(),
                 gru_layer_cls=ly.GRULayer,
                 output_nonlinearity=None,
                 input_var=None,
                 input_layer=None,
                 layer_args=None):
        with tf.compat.v1.variable_scope(name, 'GRUNetwork'):
            if input_layer is None:
                l_in = ly.InputLayer(
                    shape=(None, None) + input_shape,
                    input_var=input_var,
                    name='input')
            else:
                l_in = input_layer
            l_step_input = ly.InputLayer(
                shape=(None, ) + input_shape, name='step_input')
            l_step_prev_state = ly.InputLayer(
                shape=(None, hidden_dim), name='step_prev_state')
            if layer_args is None:
                layer_args = dict()
            l_gru = gru_layer_cls(
                l_in,
                num_units=hidden_dim,
                hidden_nonlinearity=hidden_nonlinearity,
                gate_nonlinearity=recurrent_nonlinearity,
                hidden_init_trainable=False,
                w_x_init=recurrent_w_x_init,
                w_h_init=recurrent_w_h_init,
                name='gru',
                **layer_args)
            l_gru_flat = ly.ReshapeLayer(
                l_gru, shape=(-1, hidden_dim), name='gru_flat')
            l_output_flat = ly.DenseLayer(
                l_gru_flat,
                num_units=output_dim,
                nonlinearity=output_nonlinearity,
                w=output_w_init,
                name='output_flat')
            l_output = ly.OpLayer(
                l_output_flat,
                op=lambda flat_output, l_input: tf.reshape(
                    flat_output,
                    tf.stack((tf.shape(l_input)[0], tf.shape(l_input)[1], -1))
                ),
                shape_op=lambda flat_output_shape, l_input_shape: (
                    l_input_shape[0], l_input_shape[1], flat_output_shape[-1]),
                extras=[l_in],
                name='output')
            l_step_state = l_gru.get_step_layer(
                l_step_input, l_step_prev_state, name='step_state')
            l_step_hidden = l_step_state
            l_step_output = ly.DenseLayer(
                l_step_hidden,
                num_units=output_dim,
                nonlinearity=output_nonlinearity,
                w=l_output_flat.w,
                b=l_output_flat.b,
                name='step_output')

            self._l_in = l_in
            self._hid_init_param = l_gru.h0
            self._l_gru = l_gru
            self._l_out = l_output
            self._l_step_input = l_step_input
            self._l_step_prev_state = l_step_prev_state
            self._l_step_hidden = l_step_hidden
            self._l_step_state = l_step_state
            self._l_step_output = l_step_output
            self._hidden_dim = hidden_dim

    @property
    def state_dim(self):
        return self._hidden_dim

    @property
    def hidden_dim(self):
        return self._hidden_dim

    @property
    def input_layer(self):
        return self._l_in

    @property
    def input_var(self):
        return self._l_in.input_var

    @property
    def output_layer(self):
        return self._l_out

    @property
    def recurrent_layer(self):
        return self._l_gru

    @property
    def step_input_layer(self):
        return self._l_step_input

    @property
    def step_prev_state_layer(self):
        return self._l_step_prev_state

    @property
    def step_hidden_layer(self):
        return self._l_step_hidden

    @property
    def step_state_layer(self):
        return self._l_step_state

    @property
    def step_output_layer(self):
        return self._l_step_output

    @property
    def hid_init_param(self):
        return self._hid_init_param

    @property
    def state_init_param(self):
        return self._hid_init_param


class LSTMNetwork:
    def __init__(self,
                 input_shape,
                 output_dim,
                 hidden_dim,
                 name=None,
                 hidden_nonlinearity=tf.nn.relu,
                 output_w_init=ly.XavierUniformInitializer(),
                 recurrent_nonlinearity=tf.nn.sigmoid,
                 recurrent_w_x_init=ly.XavierUniformInitializer(),
                 recurrent_w_h_init=ly.OrthogonalInitializer(),
                 lstm_layer_cls=ly.LSTMLayer,
                 output_nonlinearity=None,
                 input_var=None,
                 input_layer=None,
                 forget_bias=1.0,
                 use_peepholes=False,
                 layer_args=None):
        with tf.compat.v1.variable_scope(name, 'LSTMNetwork'):
            if input_layer is None:
                l_in = ly.InputLayer(
                    shape=(None, None) + input_shape,
                    input_var=input_var,
                    name='input')
            else:
                l_in = input_layer
            l_step_input = ly.InputLayer(
                shape=(None, ) + input_shape, name='step_input')
            # contains previous hidden and cell state
            l_step_prev_state = ly.InputLayer(
                shape=(None, hidden_dim * 2), name='step_prev_state')
            if layer_args is None:
                layer_args = dict()
            l_lstm = lstm_layer_cls(
                l_in,
                num_units=hidden_dim,
                hidden_nonlinearity=hidden_nonlinearity,
                gate_nonlinearity=recurrent_nonlinearity,
                hidden_init_trainable=False,
                name='lstm_layer',
                forget_bias=forget_bias,
                cell_init_trainable=False,
                w_x_init=recurrent_w_x_init,
                w_h_init=recurrent_w_h_init,
                use_peepholes=use_peepholes,
                **layer_args)
            l_lstm_flat = ly.ReshapeLayer(
                l_lstm, shape=(-1, hidden_dim), name='lstm_flat')
            l_output_flat = ly.DenseLayer(
                l_lstm_flat,
                num_units=output_dim,
                nonlinearity=output_nonlinearity,
                w=output_w_init,
                name='output_flat')
            l_output = ly.OpLayer(
                l_output_flat,
                op=lambda flat_output, l_input: tf.reshape(
                    flat_output,
                    tf.stack((tf.shape(l_input)[0], tf.shape(l_input)[1], -1))
                ),
                shape_op=lambda flat_output_shape, l_input_shape: (
                    l_input_shape[0], l_input_shape[1], flat_output_shape[-1]),
                extras=[l_in],
                name='output')
            l_step_state = l_lstm.get_step_layer(
                l_step_input, l_step_prev_state, name='step_state')
            l_step_hidden = ly.SliceLayer(
                l_step_state, indices=slice(hidden_dim), name='step_hidden')
            l_step_cell = ly.SliceLayer(
                l_step_state,
                indices=slice(hidden_dim, None),
                name='step_cell')
            l_step_output = ly.DenseLayer(
                l_step_hidden,
                num_units=output_dim,
                nonlinearity=output_nonlinearity,
                w=l_output_flat.w,
                b=l_output_flat.b,
                name='step_output')

            self._l_in = l_in
            self._hid_init_param = l_lstm.h0
            self._cell_init_param = l_lstm.c0
            self._l_lstm = l_lstm
            self._l_out = l_output
            self._l_step_input = l_step_input
            self._l_step_prev_state = l_step_prev_state
            self._l_step_hidden = l_step_hidden
            self._l_step_cell = l_step_cell
            self._l_step_state = l_step_state
            self._l_step_output = l_step_output
            self._hidden_dim = hidden_dim

    @property
    def state_dim(self):
        return self._hidden_dim * 2

    @property
    def input_layer(self):
        return self._l_in

    @property
    def input_var(self):
        return self._l_in.input_var

    @property
    def output_layer(self):
        return self._l_out

    @property
    def recurrent_layer(self):
        return self._l_lstm

    @property
    def step_input_layer(self):
        return self._l_step_input

    @property
    def step_prev_state_layer(self):
        return self._l_step_prev_state

    @property
    def step_hidden_layer(self):
        return self._l_step_hidden

    @property
    def step_state_layer(self):
        return self._l_step_state

    @property
    def step_cell_layer(self):
        return self._l_step_cell

    @property
    def step_output_layer(self):
        return self._l_step_output

    @property
    def hid_init_param(self):
        return self._hid_init_param

    @property
    def cell_init_param(self):
        return self._cell_init_param

    @property
    def state_init_param(self):
        return tf.concat(
            axis=0, values=[self._hid_init_param, self._cell_init_param])


class ConvMergeNetwork(LayersPowered, Serializable):
    """
    This network allows the input to consist of a convolution-friendly
    component, plus a non-convolution-friendly component. These two components
    will be concatenated in the fully connected layers. There can also be a
    list of optional layers for the non-convolution-friendly component alone.


    The input to the network should be a matrix where each row is a single
    input entry, with both the aforementioned components flattened out and then
    concatenated together
    """

    def __init__(self,
                 input_shape,
                 extra_input_shape,
                 output_dim,
                 hidden_sizes,
                 conv_filters,
                 conv_filter_sizes,
                 conv_strides,
                 conv_pads,
                 name=None,
                 extra_hidden_sizes=None,
                 hidden_w_init=ly.XavierUniformInitializer(),
                 hidden_b_init=tf.zeros_initializer(),
                 output_w_init=ly.XavierUniformInitializer(),
                 output_b_init=tf.zeros_initializer(),
                 hidden_nonlinearity=tf.nn.relu,
                 output_nonlinearity=None,
                 input_var=None,
                 input_layer=None):
        Serializable.quick_init(self, locals())

        if extra_hidden_sizes is None:
            extra_hidden_sizes = []

        with tf.compat.v1.variable_scope(name, 'ConvMergeNetwork'):

            input_flat_dim = np.prod(input_shape)
            extra_input_flat_dim = np.prod(extra_input_shape)
            total_input_flat_dim = input_flat_dim + extra_input_flat_dim

            if input_layer is None:
                l_in = ly.InputLayer(
                    shape=(None, total_input_flat_dim),
                    input_var=input_var,
                    name='input')
            else:
                l_in = input_layer

            l_conv_in = ly.reshape(
                ly.SliceLayer(
                    l_in, indices=slice(input_flat_dim), name='conv_slice'),
                ([0], ) + input_shape,
                name='conv_reshaped')
            l_extra_in = ly.reshape(
                ly.SliceLayer(
                    l_in,
                    indices=slice(input_flat_dim, None),
                    name='extra_slice'), ([0], ) + extra_input_shape,
                name='extra_reshaped')

            l_conv_hid = l_conv_in
            for idx, conv_filter, filter_size, stride, pad in zip(
                    range(len(conv_filters)),
                    conv_filters,
                    conv_filter_sizes,
                    conv_strides,
                    conv_pads,
            ):
                l_conv_hid = ly.Conv2DLayer(
                    l_conv_hid,
                    num_filters=conv_filter,
                    filter_size=filter_size,
                    stride=(stride, stride),
                    pad=pad,
                    nonlinearity=hidden_nonlinearity,
                    name='conv_hidden_%d' % idx,
                )

            l_extra_hid = l_extra_in
            for idx, hidden_size in enumerate(extra_hidden_sizes):
                l_extra_hid = ly.DenseLayer(
                    l_extra_hid,
                    num_units=hidden_size,
                    nonlinearity=hidden_nonlinearity,
                    name='extra_hidden_%d' % idx,
                    w=hidden_w_init,
                    b=hidden_b_init,
                )

            l_joint_hid = ly.concat(
                [ly.flatten(l_conv_hid, name='conv_hidden_flat'), l_extra_hid],
                name='joint_hidden')

            for idx, hidden_size in enumerate(hidden_sizes):
                l_joint_hid = ly.DenseLayer(
                    l_joint_hid,
                    num_units=hidden_size,
                    nonlinearity=hidden_nonlinearity,
                    name='joint_hidden_%d' % idx,
                    w=hidden_w_init,
                    b=hidden_b_init,
                )
            l_out = ly.DenseLayer(
                l_joint_hid,
                num_units=output_dim,
                nonlinearity=output_nonlinearity,
                name='output',
                w=output_w_init,
                b=output_b_init,
            )
            self._l_in = l_in
            self._l_out = l_out

            LayersPowered.__init__(self, [l_out], input_layers=[l_in])

    @property
    def input_layer(self):
        return self._l_in

    @property
    def output_layer(self):
        return self._l_out

    @property
    def input_var(self):
        return self._l_in.input_var



import tensorflow as tf

from garage.tf.core import Parameterized


class QFunction(Parameterized):
    def build_net(self, name):
        raise NotImplementedError

    def get_qval_sym(self, input_phs):
        raise NotImplementedError

    def log_diagnostics(self, paths):
        pass

    def get_trainable_vars(self, scope=None):
        scope = scope if scope else self.name
        return tf.compat.v1.get_collection(
            tf.compat.v1.GraphKeys.TRAINABLE_VARIABLES, scope=scope)

    def get_global_vars(self, scope=None):
        scope = scope if scope else self.name
        return tf.compat.v1.get_collection(
            tf.compat.v1.GraphKeys.GLOBAL_VARIABLES, scope=scope)

    def get_regularizable_vars(self, scope=None):
        scope = scope if scope else self.name
        reg_vars = [
            var for var in self.get_trainable_vars(scope=scope)
            if 'W' in var.name and 'output' not in var.name
        ]
        return reg_vars



#!/usr/bin/env python3
"""Example of how to load, step, and visualize an environment.

This example requires that garage[dm_control] be installed.
"""
from garage.envs.dm_control import DmControlEnv

# Construct the environment
env = DmControlEnv.from_suite('walker', 'run')

# Reset the environment and launch the viewer
env.reset()
env.render()

# Step randomly until interrupted
try:
    print('Press Ctrl-C to stop...')
    while True:
        env.step(env.action_space.sample())
        env.render()
except KeyboardInterrupt:
    print('Exiting...')
    env.close()



from dowel import logger
import numpy as np
import pytest

from garage.np.baselines import LinearFeatureBaseline
from garage.sampler.utils import truncate_paths
from garage.tf.algos import VPG
from garage.tf.envs import TfEnv
from garage.tf.experiment import LocalTFRunner
from garage.tf.policies import CategoricalMLPPolicy
from garage.tf.samplers import BatchSampler
from tests.fixtures import snapshot_config
from tests.fixtures.logger import NullOutput


class TestSampler:

    def setup_method(self):
        logger.add_output(NullOutput())

    def teardown_method(self):
        logger.remove_all()

    def test_truncate_paths(self):
        paths = [
            dict(
                observations=np.zeros((100, 1)),
                actions=np.zeros((100, 1)),
                rewards=np.zeros(100),
                env_infos=dict(),
                agent_infos=dict(lala=np.zeros(100)),
            ),
            dict(
                observations=np.zeros((50, 1)),
                actions=np.zeros((50, 1)),
                rewards=np.zeros(50),
                env_infos=dict(),
                agent_infos=dict(lala=np.zeros(50)),
            ),
        ]

        truncated = truncate_paths(paths, 130)
        assert len(truncated) == 2
        assert len(truncated[-1]['observations']) == 30
        assert len(truncated[0]['observations']) == 100
        # make sure not to change the original one
        assert len(paths) == 2
        assert len(paths[-1]['observations']) == 50

    # Note:
    #   test_batch_sampler should pass if tested independently
    #   from other tests, but cannot be tested on CI.
    #
    #   This is because nose2 runs all tests in a single process,
    #   when this test is running, tensorflow has already been initialized,
    #   and later singleton_pool will hang because tensorflow is not fork-safe.
    @pytest.mark.flaky
    def test_tf_batch_sampler(self):
        max_cpus = 8
        with LocalTFRunner(snapshot_config, max_cpus=max_cpus) as runner:
            env = TfEnv(env_name='CartPole-v1')

            policy = CategoricalMLPPolicy(name='policy',
                                          env_spec=env.spec,
                                          hidden_sizes=(32, 32))

            baseline = LinearFeatureBaseline(env_spec=env.spec)

            algo = VPG(env_spec=env.spec,
                       policy=policy,
                       baseline=baseline,
                       max_path_length=1,
                       discount=0.99)

            runner.setup(algo,
                         env,
                         sampler_cls=BatchSampler,
                         sampler_args={'n_envs': max_cpus})

            try:
                runner.initialize_tf_vars()
            except BaseException:
                raise AssertionError(
                    'LocalRunner should be able to initialize tf variables.')

            runner._start_worker()

            paths = runner.sampler.obtain_samples(0,
                                                  batch_size=8,
                                                  whole_paths=True)
            assert len(paths) >= max_cpus, (
                'BatchSampler should sample more than max_cpus={} '
                'trajectories'.format(max_cpus))



"""
This modules creates a continuous MLP policy network.

A continuous MLP network can be used as policy method in different RL
algorithms. It accepts an observation of the environment and predicts an
action.
"""
import akro
import tensorflow as tf

from garage.core import Serializable
from garage.misc.overrides import overrides
from garage.tf.core import layers
from garage.tf.core import LayersPowered
from garage.tf.core.layers import batch_norm
from garage.tf.misc import tensor_utils
from garage.tf.policies.base import Policy


class ContinuousMLPPolicy(Policy, LayersPowered, Serializable):
    """
    This class implements a policy network.

    The policy network selects action based on the state of the environment.
    It uses neural nets to fit the function of pi(s).
    """

    def __init__(self,
                 env_spec,
                 hidden_sizes=(64, 64),
                 name='ContinuousMLPPolicy',
                 hidden_nonlinearity=tf.nn.relu,
                 output_nonlinearity=tf.nn.tanh,
                 input_include_goal=False,
                 bn=False):
        """
        Initialize class with multiple attributes.

        Args:
            env_spec():
            hidden_sizes(list or tuple, optional):
                A list of numbers of hidden units for all hidden layers.
            name(str, optional):
                A str contains the name of the policy.
            hidden_nonlinearity(optional):
                An activation shared by all fc layers.
            output_nonlinearity(optional):
                An activation used by the output layer.
            bn(bool, optional):
                A bool to indicate whether normalize the layer or not.
        """
        assert isinstance(env_spec.action_space, akro.Box)

        Serializable.quick_init(self, locals())
        super(ContinuousMLPPolicy, self).__init__(env_spec)

        self.name = name
        self._env_spec = env_spec
        if input_include_goal:
            self._obs_dim = env_spec.observation_space.flat_dim_with_keys(
                ['observation', 'desired_goal'])
        else:
            self._obs_dim = env_spec.observation_space.flat_dim
        self._action_dim = env_spec.action_space.flat_dim
        self._action_bound = env_spec.action_space.high
        self._hidden_sizes = hidden_sizes
        self._hidden_nonlinearity = hidden_nonlinearity
        self._output_nonlinearity = output_nonlinearity
        self._batch_norm = bn
        self._policy_network_name = 'policy_network'
        # Build the network and initialized as Parameterized
        self._f_prob_online, self._output_layer, self._obs_layer = self.build_net(  # noqa: E501
            name=self.name)
        LayersPowered.__init__(self, [self._output_layer])

    def build_net(self, trainable=True, name=None):
        """
        Set up q network based on class attributes.

        This function uses layers defined in garage.tf.

        Args:
            reuse: A bool indicates whether reuse variables in the same scope.
            trainable: A bool indicates whether variables are trainable.
        """
        with tf.compat.v1.variable_scope(name):
            l_in = layers.InputLayer(shape=(None, self._obs_dim), name='obs')

            l_hidden = l_in
            for idx, hidden_size in enumerate(self._hidden_sizes):
                if self._batch_norm:
                    l_hidden = batch_norm(l_hidden)

                l_hidden = layers.DenseLayer(
                    l_hidden,
                    hidden_size,
                    nonlinearity=self._hidden_nonlinearity,
                    trainable=trainable,
                    name='hidden_%d' % idx)

            l_output = layers.DenseLayer(
                l_hidden,
                self._action_dim,
                nonlinearity=self._output_nonlinearity,
                trainable=trainable,
                name='output')

            with tf.name_scope(self._policy_network_name):
                action = layers.get_output(l_output)
                scaled_action = tf.multiply(
                    action, self._action_bound, name='scaled_action')

        f_prob_online = tensor_utils.compile_function(
            inputs=[l_in.input_var], outputs=scaled_action)
        output_layer = l_output
        obs_layer = l_in

        return f_prob_online, output_layer, obs_layer

    def get_action_sym(self, obs_var, name=None, **kwargs):
        """Return action sym according to obs_var."""
        with tf.name_scope(name, 'get_action_sym', [obs_var]):
            with tf.name_scope(self._policy_network_name):
                actions = layers.get_output(
                    self._output_layer, {self._obs_layer: obs_var}, **kwargs)
            return tf.multiply(actions, self._action_bound)

    @overrides
    def get_action(self, observation):
        """Return a single action."""
        return self._f_prob_online([observation])[0], dict()

    @overrides
    def get_actions(self, observations):
        """Return multiple actions."""
        return self._f_prob_online(observations), dict()

    @property
    def vectorized(self):
        return True

    def log_diagnostics(self, paths):
        pass

    def get_trainable_vars(self, scope=None):
        scope = scope if scope else self.name
        return tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope=scope)

    def get_global_vars(self, scope=None):
        scope = scope if scope else self.name
        return tf.get_collection(tf.GraphKeys.GLOBAL_VARIABLES, scope=scope)

    def get_regularizable_vars(self, scope=None):
        scope = scope if scope else self.name
        reg_vars = [
            var for var in self.get_trainable_vars(scope=scope)
            if 'W' in var.name and 'output' not in var.name
        ]
        return reg_vars



import tensorflow as tf

from garage.core import Serializable
from garage.tf.core.parameterized import Parameterized, suppress_params_loading


class Simple(Parameterized, Serializable):
    def __init__(self, name):
        Serializable.quick_init(self, locals())
        with tf.compat.v1.variable_scope(name):
            self.w = tf.compat.v1.get_variable('w', [10, 10])

    def get_params_internal(self, **tags):
        return [self.w]


class AllArgs(Serializable):
    def __init__(self, vararg, *args, **kwargs):
        Serializable.quick_init(self, locals())
        self.vararg = vararg
        self.args = args
        self.kwargs = kwargs


class TestSerializable:
    def test_serializable(self):
        with suppress_params_loading():
            obj = Simple(name='obj')
            obj1 = Serializable.clone(obj, name='obj1')
            assert obj.w.name.startswith('obj/')
            assert obj1.w.name.startswith('obj1/')

            obj2 = AllArgs(0, *(1, ), **{'kwarg': 2})
            obj3 = Serializable.clone(obj2)
            assert obj3.vararg == 0
            assert len(obj3.args) == 1 and obj3.args[0] == 1
            assert len(obj3.kwargs) == 1 and obj3.kwargs['kwarg'] == 2



import gym


class DummyEnv(gym.Env):
    """Base dummy environment."""

    def __init__(self, random, obs_dim=(4, ), action_dim=(2, )):
        self.random = random
        self.state = None
        self._obs_dim = obs_dim
        self._action_dim = action_dim

    @property
    def observation_space(self):
        """Return an observation space."""
        raise NotImplementedError

    @property
    def action_space(self):
        """Return an action space."""
        raise NotImplementedError

    def reset(self):
        """Reset the environment."""
        raise NotImplementedError

    def step(self, action):
        """Step the environment."""
        raise NotImplementedError



"""
This script creates a test that fails when
garage.tf.baselines failed to initialize.
"""
import tensorflow as tf

from garage.envs.wrappers import Resize
from garage.tf.baselines import ContinuousMLPBaseline
from garage.tf.baselines import GaussianConvBaseline
from garage.tf.baselines import GaussianMLPBaseline
from garage.tf.envs import TfEnv
from tests.fixtures import TfGraphTestCase
from tests.fixtures.envs.dummy import DummyBoxEnv
from tests.fixtures.envs.dummy import DummyDiscrete2DEnv


class TestTfBaselines(TfGraphTestCase):
    def test_baseline(self):
        """Test the baseline initialization."""
        box_env = TfEnv(DummyBoxEnv())
        deterministic_mlp_baseline = ContinuousMLPBaseline(env_spec=box_env)
        gaussian_mlp_baseline = GaussianMLPBaseline(env_spec=box_env)

        discrete_env = TfEnv(Resize(DummyDiscrete2DEnv(), width=64, height=64))
        gaussian_conv_baseline = GaussianConvBaseline(
            env_spec=discrete_env,
            regressor_args=dict(
                conv_filters=[32, 32],
                conv_filter_sizes=[1, 1],
                conv_strides=[1, 1],
                conv_pads=['VALID', 'VALID'],
                hidden_sizes=(32, 32)))

        self.sess.run(tf.compat.v1.global_variables_initializer())
        deterministic_mlp_baseline.get_param_values(trainable=True)
        gaussian_mlp_baseline.get_param_values(trainable=True)
        gaussian_conv_baseline.get_param_values(trainable=True)

        box_env.close()
